/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.syncdocs;

import java.util.HashMap;

public class AppNameMapping {

	private static final Integer ONE = new Integer(1);

	private HashMap appIdToAppNameMap;
	private HashMap appNameCollisionMap;

	public AppNameMapping() {
		appIdToAppNameMap = new HashMap();
		appNameCollisionMap = new HashMap();
	}

	public void addMapping(String appId, String appName) {
		appIdToAppNameMap.put(appId, appName);
		increment(appNameCollisionMap, appName);
	}

	public String getNameForId(String appId) {
		String appName = (String) appIdToAppNameMap.get(appId);

		if (hasCollisions(appNameCollisionMap, appName)) {
			return appName + " (" + appId + ")"; 
		}

		return appName;
	}

	private void increment(HashMap collisionMap, String appName) {
		Integer collisions = (Integer) collisionMap.get(appName);
		if (collisions == null) {
			collisionMap.put(appName, ONE);
		}
		else {
			collisionMap.put(appName, new Integer(collisions.intValue() + 1));
		}
	}

	private boolean hasCollisions(HashMap collisionMap, String appName) {
		Integer collisions = (Integer) collisionMap.get(appName);
		if (collisions == null || collisions.intValue() < 1) {
			throw new IllegalStateException("Map not correctly set up.");
		}
		else {
			return (collisions.intValue() > 1);
		}
	}

}
