/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.syncdocs;

import java.io.IOException;
import java.net.InetAddress;
import java.util.List;

import com.apple.dnssd.DNSSD;
import com.apple.dnssd.DNSSDException;
import com.apple.dnssd.DNSSDRegistration;
import com.apple.dnssd.DNSSDService;
import com.apple.dnssd.RegisterListener;

public class BonjourThread extends ZeroConfThread {

	private RegisterListener registerListener = null;
	private DNSSDRegistration dnssdRegistration = null;

	public BonjourThread(InetAddress netAddress, String serviceName, int port,
			List zeroConfThreads, IExceptionCallback callback,
			ITerminatedCallback termCallback)
	{
		super(netAddress, serviceName, port, zeroConfThreads, callback, termCallback);

		registerListener = new BonjourRegisterListener();
	}

	public static boolean requiresOneThreadPerDomain() {
		return false;
	}

	public void bind() throws IOException {
		try {
			dnssdRegistration = DNSSD.register(serviceName + "._http._tcp.local.", "_http._tcp.",
					port, registerListener);

			System.out.println("Bonjour " + Thread.currentThread().getName() + ":  " + iNetAddress + " started at " + System.currentTimeMillis());
		}
		catch (DNSSDException e) {
			e.printStackTrace();
			throw new IOException("Unable to bind to Bonjour, underlying cause:  '" + e.getMessage() + "'.");
		}
	}

	public void protectedUnbind() {
		if (dnssdRegistration != null) {
			dnssdRegistration.stop();
		}
	}

	public class BonjourRegisterListener implements RegisterListener {

		public void serviceRegistered(DNSSDRegistration arg0, int arg1,
				String arg2, String arg3, String arg4)
		{
			System.out.println("Bonjour successfully registered at " + System.currentTimeMillis());
		}

		public void operationFailed(DNSSDService arg0, int arg1) {
			System.out.println("Failed to bind to Bonjour at " + System.currentTimeMillis());
			exCallback.handleException(new IOException("Failed to bind to Bonjour."));
		}

	}

	public static boolean isAvailable() {
		try {
			Class.forName("com.apple.dnssd.DNSSD");
			System.out.println("Apple's Bonjour implementation is available.");
			return true;
		}
		catch (ClassNotFoundException e) {
			System.out.println("Apple's Bonjour implementation is not available.");
			return false;
		}
		catch (UnsatisfiedLinkError e) {
			System.out.println("Apple's Bonjour library is not available:  '" + e.getMessage() + "'.");
			return false;
		}
	}

}
