/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.syncdocs;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

public class Messages {
	private static final String BUNDLE_NAME = "ca.quine.syncdocs.messages"; //$NON-NLS-1$

	private List listeners;

	private ResourceBundle resourceBundle = ResourceBundle.getBundle(BUNDLE_NAME);

	private static final Locale[] AVAILABLE_LOCALES = new Locale[] {
			Locale.ENGLISH, Locale.GERMAN, new Locale("pt", "PT") };

	public Messages() {
		listeners = new ArrayList();
		System.out.println("Using locale '" + resourceBundle.getLocale() + "'.");
	}

	public Locale getCurrentLocale() {
		return resourceBundle.getLocale();
	}

	public void setLocale(Locale selectedLocale) {
		PropertyChangeEvent event = new PropertyChangeEvent(this, "locale",
				getCurrentLocale(), selectedLocale);

		Locale.setDefault(selectedLocale);
		resourceBundle = ResourceBundle.getBundle(BUNDLE_NAME);

		Iterator listenerIter = listeners.iterator();
		while (listenerIter.hasNext()) {
			PropertyChangeListener listener = (PropertyChangeListener) listenerIter.next();
			listener.propertyChange(event);
		}
	}

	public Locale[] getAvailableLocales() {
		return AVAILABLE_LOCALES;
	}

	public String getString(String key) {
		try {
			return resourceBundle.getString(key);
		}
		catch (MissingResourceException e) {
			return '!' + key + '!';
		}
	}

	public void registerListener(PropertyChangeListener listener) {
		listeners.add(listener);
	}

	public void unregisterListener(PropertyChangeListener listener) {
		listeners.remove(listener);
	}

}
