/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.syncdocs;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.TransferHandler;

public class MyTransferHandler extends TransferHandler
{
	private static final long serialVersionUID = -929256731543735778L;

	private List/* <FileReceivedHandler> */handlerList;

	public MyTransferHandler() {
		handlerList = new ArrayList();
	}

	public boolean canImport(JComponent cp, DataFlavor[] df) {
		for (int i = 0; i < df.length; i++) {
			if (df[i].equals(DataFlavor.javaFileListFlavor)) {
				return true;
			}

			if (df[i].equals(DataFlavor.stringFlavor)) {
				return true;
			}

			if (uriListMimeType(df[i].getMimeType()) &&
					df[i].getRepresentationClass().equals(String.class))
			{
				return true;
			}
		}

		System.out.println("Import request has no recognized data flavour.");

		for (int i = 0; i < df.length; i++) {
			System.out.println("Data flavour " + i + ":");
			System.out.println("MIME type '" + df[i].getMimeType() + "'");
			System.out.println("Representation class '" + df[i].getRepresentationClass() + "'");
			System.out.println("Default representation class '" + df[i].getDefaultRepresentationClass() + "'");
			System.out.println("Default representation class as String '" + df[i].getDefaultRepresentationClassAsString() + "'");
		}

		return false;
	}

	private boolean uriListMimeType(String mimeType) {
		return mimeType.startsWith("text/uri-list;") ||
				mimeType.equals("text/uri-list");
	}

	private boolean hasFileFlavor(DataFlavor[] df) {
		boolean result = false;
		for (int i = 0; i < df.length; i++) {
			DataFlavor flavour = df[i];

			result = DataFlavor.javaFileListFlavor.equals(flavour);

			if (result) {
				break;
			}
		}

		return result;
	}

	private boolean hasStringFlavor(DataFlavor[] df) {
		boolean result = false;
		for (int i = 0; i < df.length; i++) {
			DataFlavor flavour = df[i];

			result = DataFlavor.stringFlavor.equals(flavour);
			if (result) {
				break;
			}
		}

		return result;
	}

	private DataFlavor getUriListFlavor(DataFlavor[] df) {
		for (int i = 0; i < df.length; i++) {
			if (uriListMimeType(df[i].getMimeType()) &&
					df[i].getRepresentationClass().equals(String.class))
			{
				return df[i];
			}
		}

		return null;
	}

	public boolean importData(JComponent cp, Transferable tr) {
		if (this.hasFileFlavor(tr.getTransferDataFlavors())) {
			try {
				List files = (List)
						(tr.getTransferData(DataFlavor.javaFileListFlavor));

				for (int i = 0; i < files.size(); i++) {
					File file = (File) files.get(i);
					notify(file.getAbsolutePath());
				}

				return true;
			}
			catch (Exception e) {
				e.printStackTrace();
			}

			return false;
		}

		if (this.hasStringFlavor(tr.getTransferDataFlavors())) {
			try {
				String myData = (String)
						(tr.getTransferData(DataFlavor.stringFlavor));

				handleUriList(myData, tr.getTransferDataFlavors());

				return true;
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}

		DataFlavor uriListFlavor =
				this.getUriListFlavor(tr.getTransferDataFlavors());

		if (uriListFlavor != null) {
			try {
				String uriList = (String) tr.getTransferData(uriListFlavor);

				handleUriList(uriList, tr.getTransferDataFlavors());

				return true;
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}

		DataFlavor[] transferDataFlavors = tr.getTransferDataFlavors();
		for (int i = 0; i < transferDataFlavors.length; i++) {
			DataFlavor flavour = transferDataFlavors[i];
			System.out.println("Unsupported flavour '" + flavour + "'.");
		}

		return false;
	}

	private void handleUriList(String uriList, DataFlavor[] flavors) throws MalformedURLException {
		if (uriList.startsWith("file:")) {
			String[] filenames = uriList.split("\n");

			for (int i = 0; i < filenames.length; i++) {
				URL fileUrl = new URL(filenames[i]);
				File file = new File(fileUrl.getPath());
				notify(file.getAbsolutePath());
			}
		}
		else {
			System.out.println("Received a String as a drop source:  '" + uriList + "', but it is not a URL.");
			System.out.println("The MIME types are:  (" + dataFlavorsToMimeTypeString(flavors) + ").");
		}
	}

	private String dataFlavorsToMimeTypeString(DataFlavor[] flavors) {
		StringBuffer sb = new StringBuffer();

		for (int i = 0; i < flavors.length; i++) {
			DataFlavor flavor = flavors[i];
			sb.append(flavor.getMimeType());
			if (i < flavors.length) {
				sb.append(", ");
			}
		}

		return sb.toString();
	}

	private void notify(String filename) {
		Iterator handlerIter = handlerList.iterator();
		while (handlerIter.hasNext()) {
			FileReceivedHandler handler =
					(FileReceivedHandler) handlerIter.next();

			handler.receiveFileName(filename);
		}
	}

	public void addFileReceivedHandler(FileReceivedHandler handler) {
		handlerList.add(handler);
	}

	public void removeFileReceivedHandler(FileReceivedHandler handler) {
		handlerList.remove(handler);
	}

}
