/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.syncdocs;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.InputMap;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;

public class NetworkInterfaceDialog extends JDialog {

	private static final long serialVersionUID = 8581099319295878798L;

	private JPanel centrePanel = null;
	private JComboBox networkInterfaceBox = null;
	private InetAddress selectedAddress = null;

	public NetworkInterfaceDialog(final JFrame frame, final Messages messages,
			String title, boolean allowCancel, List networkDetailList)
	{
		super(frame, title, true);

		if (!allowCancel) {
			// Prevent the user from closing the dialog with the "X" button.
			setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
		}

		centrePanel = new JPanel();
		centrePanel.setLayout(new BorderLayout());

		getContentPane().setLayout(new BorderLayout());
		getContentPane().add(centrePanel, BorderLayout.CENTER);

		JTextArea directions = new JTextArea(3, 20);
		directions.setLineWrap(true);
		directions.setWrapStyleWord(true);
		directions.setEditable(false);
		// This computer has multiple network addresses (normally due to having both a wired and wireless connection).
		directions.setText(messages.getString("NetworkInterfaceDialog.label.whichIsWirelessInterface")); //$NON-NLS-1$
		directions.setOpaque(false);
		centrePanel.add(directions, BorderLayout.NORTH);

		networkInterfaceBox = new JComboBox();
		fillNetworkInterfaceBox(networkDetailList);
		centrePanel.add(networkInterfaceBox, BorderLayout.CENTER);

		JTextArea directions2 = new JTextArea(3, 20);
		directions2.setLineWrap(true);
		directions2.setWrapStyleWord(true);
		directions2.setEditable(false);
		directions2.setText(messages.getString("NetworkInterfaceDialog.label.dontRecognizeInterface"));  //$NON-NLS-1$
		directions2.setOpaque(false);
		centrePanel.add(directions2, BorderLayout.SOUTH);

		JPanel buttonPanel = new JPanel();
		buttonPanel.setLayout(new FlowLayout(FlowLayout.RIGHT));

		JButton refreshButton = new JButton();
		refreshButton.setAction(new AbstractAction() {
			private static final long serialVersionUID = 2066690491974966495L;

			public void actionPerformed(ActionEvent arg0) {
				try {
					List ifaceList = getAllNetworkInterfaces();
					List filteredIfaceList = buildInterfaceList(ifaceList);
					networkInterfaceBox.removeAllItems();
					fillNetworkInterfaceBox(filteredIfaceList);
				}
				catch (SocketException e) {
					JOptionPane.showMessageDialog(frame,
							messages.getString("NetworkInterfaceDialog.label.unableToRefreshInterfaceList"), //$NON-NLS-1$
							messages.getString("NetworkInterfaceDialog.title.unableToRefreshInterfaceList"), //$NON-NLS-1$
							JOptionPane.ERROR_MESSAGE);
				}
			}

		});
		refreshButton.setText(messages.getString("NetworkInterfaceDialog.button.refresh")); //$NON-NLS-1$
		buttonPanel.add(refreshButton);

		JButton helpButton = new JButton();
		helpButton.setAction(new AbstractAction() {
			private static final long serialVersionUID = -1019466191627806969L;

			public void actionPerformed(ActionEvent arg0) {
				String url = "http://syncdocs.sf.net/network-interface-help.php"; //$NON-NLS-1$

				SyncDocs.launch(url, SyncDocs.SHOW_SITE_APP_LIST,
						messages.getString("SyncDocs.error.unableToLaunchBrowser"),
						messages.getString("SyncDocs.error.unableToLaunchBrowser2"));
			}

		});
		helpButton.setText(messages.getString("SyncDocs.button.help")); //$NON-NLS-1$
		buttonPanel.add(helpButton);

		JButton okButton = new JButton();
		JButton cancelButton = new JButton();

		final List finalList = networkDetailList;
		Action okActionListener = new AbstractAction() {
			private static final long serialVersionUID = 1936089519459429819L;

			public void actionPerformed(ActionEvent actionEvent) {
				 int index = networkInterfaceBox.getSelectedIndex();
				 NetworkDetails details = (NetworkDetails) finalList.get(index);
				 selectedAddress = details.getInetAddress();

				 setVisible(false);
			}
		};

		Action cancelActionListener = new AbstractAction() {
			private static final long serialVersionUID = 1L;

			public void actionPerformed(ActionEvent actionEvent) {
				setVisible(false);
			}
		};

		if (allowCancel) {
			InputMap inputMap = rootPane.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
		    inputMap.put(KeyStroke.getKeyStroke("ESCAPE"), "ESCAPE"); //$NON-NLS-1$ //$NON-NLS-2$
		    rootPane.getActionMap().put("ESCAPE", cancelActionListener); //$NON-NLS-1$
		}

		InputMap inputMap = rootPane.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
		inputMap.put(KeyStroke.getKeyStroke("ENTER"), "ENTER"); //$NON-NLS-1$ //$NON-NLS-2$
	    rootPane.getActionMap().put("ENTER", okActionListener); //$NON-NLS-1$

	    okButton.setAction(okActionListener);
		okButton.setText(messages.getString("SyncDocs.button.OK")); //$NON-NLS-1$

		if (allowCancel) {
			cancelButton.setAction(cancelActionListener);
			cancelButton.setText(messages.getString("SyncDocs.button.cancel")); //$NON-NLS-1$
		}

		if (System.getProperty("os.name").startsWith("Mac")) { //$NON-NLS-1$ //$NON-NLS-2$
			if (allowCancel) { buttonPanel.add(cancelButton); }
			buttonPanel.add(okButton);
		} else {
			buttonPanel.add(okButton);
			if (allowCancel) { buttonPanel.add(cancelButton); }
		}

		getContentPane().add(buttonPanel, BorderLayout.SOUTH);

		//setMinimumSize(new Dimension(400, 200));
		//setPreferredSize(new Dimension(400, 200));

		pack();
		setLocationRelativeTo(frame);
		setVisible(true);
	}

	private void fillNetworkInterfaceBox(List networkDetailList) {
		Iterator networkDetailIter = networkDetailList.iterator();
		while (networkDetailIter.hasNext()) {
			NetworkDetails details = (NetworkDetails) networkDetailIter.next();
			String detailString = details.getInterfaceName() +
					" - " + details.getInetAddressForDisplay(); //$NON-NLS-1$

			networkInterfaceBox.addItem(detailString);
		}
	}

	public static List getAllNetworkInterfaces() throws SocketException {
		List ifaceList = new ArrayList();
		Enumeration interfaceEnum = NetworkInterface.getNetworkInterfaces();
	    while (interfaceEnum.hasMoreElements()) {
	    	NetworkInterface iface = (NetworkInterface) interfaceEnum.nextElement();
			ifaceList.add(iface);
		}

	    return ifaceList;
	}

	public static List buildInterfaceList(List ifaceList) {
		List networkInterfaces = new ArrayList();

		Iterator ifaceIter = ifaceList.iterator();

		while (ifaceIter.hasNext()) {
			NetworkInterface iface = (NetworkInterface) ifaceIter.next();

			if (iface.getName().startsWith("lo")) { //$NON-NLS-1$
				continue;
			}

			StringBuffer ifaceDetails = new StringBuffer();
			ifaceDetails.append(iface.getDisplayName());

			if (!iface.getName().equals(iface.getDisplayName())) {
				ifaceDetails.append(" - " + iface.getName()); //$NON-NLS-1$
			}

			Enumeration addressEnum = iface.getInetAddresses();
			while (addressEnum.hasMoreElements()) {
				InetAddress address = (InetAddress) addressEnum.nextElement();
				if (address.getAddress().length == 4) {
					// IPv4
					NetworkDetails details = new NetworkDetails();
					details.setInetAddress(address);
					details.setInterfaceName(ifaceDetails.toString());

					StringBuffer buffer = new StringBuffer();
					convertToDotted(buffer, address.getAddress());
					details.setInetAddressForDisplay(buffer.toString());

					networkInterfaces.add(details);
				}
			}
		}

		return networkInterfaces;
	}

	public static InetAddress convertToInetAddress(String dottedQuad)
			throws UnknownHostException
	{
		byte[] addressInBytes = new byte[4];

		String[] addressParts = dottedQuad.split("\\."); //$NON-NLS-1$
		addressInBytes[0] = (byte) Integer.parseInt(addressParts[0]);
		addressInBytes[1] = (byte) Integer.parseInt(addressParts[1]);
		addressInBytes[2] = (byte) Integer.parseInt(addressParts[2]);
		addressInBytes[3] = (byte) Integer.parseInt(addressParts[3]);

		return InetAddress.getByAddress(addressInBytes);
	}

	public static void convertToDotted(StringBuffer sb, byte[] address) {
		if (address.length == 4) {
			// Handle an IPv4 address.
			for (int i = 0; i < address.length; i++) {
				byte b = address[i];
				sb.append(unsignedByteToInt(b));
				if (i < address.length - 1) {
					sb.append("."); //$NON-NLS-1$
				}
			}
		}
		else if (address.length == 16) {
			boolean collapsed = false;

			// Handle an IPv6 address.
			for (int i = 0; i < address.length; i += 2) {
				byte b1 = address[i];
				byte b2 = address[i+1];

				if (b1 == 0 && b2 == 0) {
					if (!collapsed) {
						sb.append(":"); //$NON-NLS-1$
						collapsed = true;
					}

					continue;
				}

				collapsed = false;

				int i1 = unsignedByteToInt(b1);
				int i2 = unsignedByteToInt(b2);

				int shortPart = (i1 << 8) + i2;

				sb.append(Integer.toHexString(shortPart));

				if (i < address.length - 2) {
					sb.append(":"); //$NON-NLS-1$
				}
			}
		}
	}

	private static int unsignedByteToInt(byte b) {
		return (int) b & 0xFF;
	}

	public InetAddress getSelectedAddress() {
		return selectedAddress;
	}

}
