/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.syncdocs;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.Iterator;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ButtonGroup;
import javax.swing.InputMap;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

public class SetupDialog extends JDialog {

	private static final long serialVersionUID = 2115054742771325882L;

	private static final int INSET_SIZE = 2;

	private JPanel myPanel = null;
	private JTextField portTextField = null;
	private JTextField bonjourNameTextField = null;
	private JComboBox zeroConfImplementation = null;
	private boolean okSelected = false;
	private String bindAddress = null;
	private JLabel bindAddressLabel = null;
	private boolean bindAll;
	private int checkVersionFreqInDays = 0;

	private Preferences prefs;

	public SetupDialog(final SyncDocs syncDocs, final JFrame frame,
			String title, boolean allowCancel, int recommendedPort,
			final Preferences inputPrefs)
	{
		super(frame, title, true);

		final Messages messages = syncDocs.getMessages();

		if (!allowCancel) {
			// Prevent the user from closing the dialog with the "X" button.
			setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
		}

		StringBuffer sb = new StringBuffer();
		if (inputPrefs.getBindAddress() != null) {
			NetworkInterfaceDialog.convertToDotted(sb, inputPrefs.getBindAddress().getAddress());
			bindAddress = sb.toString();
		}
		else {
			bindAddress = ""; //$NON-NLS-1$
		}
		bindAll = inputPrefs.isBindAll();

		myPanel = new JPanel();
		getContentPane().add(myPanel);
		myPanel.setLayout(new BorderLayout());

		GridBagLayout gbLayout = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();

		JPanel optionPanel = new JPanel();
		optionPanel.setLayout(gbLayout);

		c.gridx = 0;
		c.gridy = 0;
		c.anchor = GridBagConstraints.LINE_START;
		c.insets = new Insets(INSET_SIZE, INSET_SIZE, INSET_SIZE, INSET_SIZE);
		optionPanel.add(new JLabel(messages.getString("SetupDialog.label.port")), c); //$NON-NLS-1$

		portTextField = new JTextField(5);
		portTextField.setText("" + inputPrefs.getPort()); //$NON-NLS-1$
		c.gridx = 1;
		c.gridy = 0;
		optionPanel.add(portTextField, c);

		JLabel validPorts = new JLabel();
		if (System.getProperty("os.name").startsWith("Windows")) { //$NON-NLS-1$ //$NON-NLS-2$
			validPorts.setText("(1-65535)"); //$NON-NLS-1$
		}
		else {
			validPorts.setText("(1024-65535)"); //$NON-NLS-1$
		}
		c.gridx = 2;
		c.gridy = 0;
		optionPanel.add(validPorts, c);

		c.gridx = 0;
		c.gridy = 1;
		c.gridwidth = 3;
		if (recommendedPort != 0) {
			optionPanel.add(new JLabel(messages.getString("SetupDialog.label.recommendedPort1") + recommendedPort + messages.getString("SetupDialog.label.recommendedPort2")), c); //$NON-NLS-1$ //$NON-NLS-2$
		}

		c.gridx = 0;
		c.gridy = 2;
		c.gridwidth = 1;
		optionPanel.add(new JLabel(messages.getString("SetupDialog.label.bonjourName")), c); //$NON-NLS-1$

		c.gridx = 1;
		c.gridy = 2;
		c.gridwidth = 2;
		bonjourNameTextField = new JTextField(10);
		bonjourNameTextField.setText(inputPrefs.getBonjourName());
		optionPanel.add(bonjourNameTextField, c);

		final JRadioButton wirelessAddressButton = new JRadioButton();
		bindAddressLabel = new JLabel("255.255.255.255"); //$NON-NLS-1$
		final JLabel finalbindAddressLabel = bindAddressLabel; 
		final JButton changeBindAddressButton = new JButton();
		final JRadioButton bindAllButton = new JRadioButton();

		try {
			if (SyncDocs.isBonjourAvailable()) {
				c.gridx = 0;
				c.gridy++;
				c.gridwidth = 1;
				optionPanel.add(new JLabel(messages.getString("SetupDialog.label.zeroConfImplementation")), c); //$NON-NLS-1$

				c.gridx = 1;
				c.gridwidth = 2;
				zeroConfImplementation = new JComboBox();
				zeroConfImplementation.addItem("JmDNS");
				zeroConfImplementation.addItem("Bonjour");
				//bonjourNameTextField.setText(inputPrefs.getBonjourImplementation());
				optionPanel.add(zeroConfImplementation, c);
				zeroConfImplementation.setAction(new AbstractAction() {

					private static final long serialVersionUID = 1L;

					public void actionPerformed(ActionEvent arg0) {
						if (zeroConfImplementation.getSelectedIndex() == 0) {
							wirelessAddressButton.setEnabled(true);
							finalbindAddressLabel.setEnabled(true);
							changeBindAddressButton.setEnabled(true);
							bindAllButton.setEnabled(true);
						}
						else {
							wirelessAddressButton.setEnabled(false);
							finalbindAddressLabel.setEnabled(false);
							changeBindAddressButton.setEnabled(false);
							bindAllButton.setEnabled(false);
						}
					}
		
				});

				zeroConfImplementation.setSelectedItem(inputPrefs.getZeroConfImplementation());
			}
		}
		catch (NoClassDefFoundError e) {
			// Do nothing, which will prevent the display of an option the user
			// doesn't have.
		}

		ButtonGroup bindGroup = new ButtonGroup();

		c.gridx = 0;
		c.gridy++;
		c.gridwidth = 1;
		wirelessAddressButton.setAction(new AbstractAction() {
			private static final long serialVersionUID = 1L;

			public void actionPerformed(ActionEvent event) {
				bindAll = false;
			}
		});
		wirelessAddressButton.setText(messages.getString("SetupDialog.label.bindToAddress")); //$NON-NLS-1$
		if (!bindAll) {
			wirelessAddressButton.setSelected(true);
		}
		bindGroup.add(wirelessAddressButton);
		optionPanel.add(wirelessAddressButton, c);

		c.gridx = 1;
		optionPanel.add(bindAddressLabel, c);

		c.gridx = 2;
		changeBindAddressButton.setAction(new AbstractAction() {
			private static final long serialVersionUID = 4362693482610592519L;

			public void actionPerformed(ActionEvent arg0) {
				try {
					InetAddress networkInterface = null;

					List ifaceList = NetworkInterfaceDialog.getAllNetworkInterfaces();
				    List networkInterfaces = NetworkInterfaceDialog.buildInterfaceList(ifaceList);

				    // If there was a previous bindAddress set, make sure it
				    // shows up first in the list (then it will be
				    // automatically selected by default).
				    if (bindAddress != null) {
				    	Iterator networkInterfaceIter = networkInterfaces.iterator();
				    	while (networkInterfaceIter.hasNext()) {
							NetworkDetails details = (NetworkDetails) networkInterfaceIter.next();
							if (details.getInetAddressForDisplay().equals(bindAddress)) {
								networkInterfaceIter.remove();
								networkInterfaces.add(0, details);
								break;
							}
						}
				    }

				    // if (ifaceList.size() > 2) {
						networkInterface = SyncDocs.getUserChoiceOfInterface(
								syncDocs.getMessages(), frame,
								networkInterfaces, bindAddress != null);

						if (networkInterface != null) {
							StringBuffer sb = new StringBuffer();
					    	NetworkInterfaceDialog.convertToDotted(sb, networkInterface.getAddress());
					    	bindAddress = sb.toString();
					    	bindAddressLabel.setText(bindAddress);
						}
//					}
//				    else {
//						JOptionPane.showMessageDialog(frame,
//								"No other network interfaces are available.",
//								"Unable to Find Other Network Interfaces",
//								JOptionPane.ERROR_MESSAGE);
//				    }
				}
				catch (SocketException e) {
					JOptionPane.showMessageDialog(frame,
							messages.getString("SetupDialog.label.networkInterfacesNotFound"), //$NON-NLS-1$
							messages.getString("SetupDialog.title.networkInterfacesNotFound"), //$NON-NLS-1$
							JOptionPane.ERROR_MESSAGE);
				}
			}
		});
		changeBindAddressButton.setText(messages.getString("SetupDialog.button.change")); //$NON-NLS-1$
		optionPanel.add(changeBindAddressButton, c);

		c.gridx = 0;
		c.gridy++;
		bindAllButton.setAction(new AbstractAction() {
			private static final long serialVersionUID = 1L;

			public void actionPerformed(ActionEvent event) {
				try {
					List inetAddressList = SyncDocs.getAllInetAddresses();
					if (inetAddressList.size() > 3) {
						JOptionPane.showMessageDialog(frame,
								messages.getString("SetupDialog.label.bindToAllNotRecommended"), //$NON-NLS-1$
								messages.getString("SetupDialog.title.bindToAllNotRecommended"), //$NON-NLS-1$
								JOptionPane.WARNING_MESSAGE);
					}
				}
				catch (SocketException e) { e.printStackTrace(); }

				bindAll = true;
			}
		});
		bindAllButton.setText(messages.getString("SetupDialog.label.bindToAllAddresses")); //$NON-NLS-1$
		if (bindAll) {
			bindAllButton.setSelected(true);
		}
		bindGroup.add(bindAllButton);
		optionPanel.add(bindAllButton, c);

		c.gridy++;

		c.gridx = 0;
		optionPanel.add(new JLabel(messages.getString("SetupDialog.label.checkForNewVersions")), c); //$NON-NLS-1$

		ButtonGroup checkFrequencyGroup = new ButtonGroup();

		checkVersionFreqInDays = inputPrefs.getCheckVersionFreqInDays();

		c.gridx = 1;
		JRadioButton onceADay = new JRadioButton();
		onceADay.setAction(new AbstractAction() {
			private static final long serialVersionUID = 1L;

			public void actionPerformed(ActionEvent event) {
				checkVersionFreqInDays = 1;
			}
		});
		checkFrequencyGroup.add(onceADay);
		onceADay.setText(messages.getString("SetupDialog.label.daily")); //$NON-NLS-1$
		if (checkVersionFreqInDays != 7 || checkVersionFreqInDays != 0) {
			onceADay.setSelected(true);
		}
		optionPanel.add(onceADay, c);

		c.gridx = 2;
		JRadioButton onceAWeek = new JRadioButton();
		onceAWeek.setAction(new AbstractAction() {
			private static final long serialVersionUID = 1L;

			public void actionPerformed(ActionEvent event) {
				checkVersionFreqInDays = 7;
			}
		});
		checkFrequencyGroup.add(onceAWeek);
		onceAWeek.setText(messages.getString("SetupDialog.label.weekly")); //$NON-NLS-1$
		if (checkVersionFreqInDays == 7) {
			onceAWeek.setSelected(true);
		}
		optionPanel.add(onceAWeek, c);

		c.gridy++;

		c.gridx = 1;
		JRadioButton never = new JRadioButton();
		never.setAction(new AbstractAction() {
			private static final long serialVersionUID = 1L;

			public void actionPerformed(ActionEvent event) {
				checkVersionFreqInDays = 0;
			}
		});
		checkFrequencyGroup.add(never);
		never.setText(messages.getString("SetupDialog.label.never")); //$NON-NLS-1$
		if (checkVersionFreqInDays == 0) {
			never.setSelected(true);
		}
		optionPanel.add(never, c);

		c.gridx = 2;
		JButton checkNow = new JButton();
		checkNow.setAction(new AbstractAction() {
			private static final long serialVersionUID = 1L;

			public void actionPerformed(ActionEvent event) {
				Boolean newVersionAvailable = syncDocs.isNewVersionAvailable();
				if (newVersionAvailable != null) {
					if (!newVersionAvailable.booleanValue()) {
						JOptionPane.showMessageDialog(null, messages.getString("SetupDialog.label.syncDocsUpToDate")); //$NON-NLS-1$
					}
				}
			}
		});
		checkNow.setText(messages.getString("SetupDialog.button.checkNow")); //$NON-NLS-1$
		optionPanel.add(checkNow, c);

		c.gridx = 0;
		c.gridy++;
		c.gridwidth = 3;
		final JCheckBox useUserDirForDownloads =
				new JCheckBox(messages.getString("SetupDialog.label.useUserDir")); //$NON-NLS-1$

		if (inputPrefs.isUseUserDirForDownloads()) {
			useUserDirForDownloads.setSelected(true);
		}

		optionPanel.add(useUserDirForDownloads, c);
		c.gridwidth = 1;

		myPanel.add(BorderLayout.CENTER, optionPanel);

		JPanel buttonPanel = new JPanel();
		buttonPanel.setLayout(new FlowLayout(FlowLayout.RIGHT));

		JButton okButton = new JButton();
		JButton cancelButton = new JButton();

		final JFrame finalFrame = frame;
		Action okActionListener = new AbstractAction() {
			private static final long serialVersionUID = -5006811889823988629L;

			public void actionPerformed(ActionEvent actionEvent) {
				okSelected = true;

				int port = 0;
				try {
					port = Integer.parseInt(portTextField.getText());
					String bonjourName = bonjourNameTextField.getText().trim();

					if (SyncDocs.isEmpty(bonjourName)) {
						JOptionPane.showMessageDialog(finalFrame, messages.getString("SetupDialog.label.pleaseEnterBonjourName"), messages.getString("SetupDialog.title.pleaseEnterBonjourName"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$ //$NON-NLS-2$
						return;
					}
					else if (bonjourName.indexOf(".") > -1) { //$NON-NLS-1$
						JOptionPane.showMessageDialog(finalFrame,
								messages.getString("SetupDialog.label.bonjourNameCannotContainPeriod"), //$NON-NLS-1$
								messages.getString("SetupDialog.title.bonjourNameCannotContainPeriod"), //$NON-NLS-1$
								JOptionPane.ERROR_MESSAGE);

						return;
					}

					InetAddress inetAddress = null;
					try {
						if (!SyncDocs.isEmpty(bindAddress)) {
							inetAddress =
								NetworkInterfaceDialog.convertToInetAddress(bindAddress);
						}

						if (wirelessAddressButton.isSelected()) {
							if (SyncDocs.isEmpty(bindAddress)) {
								JOptionPane.showMessageDialog(frame,
										messages.getString("SetupDialog.label.setBindingAddress"), //$NON-NLS-1$
										messages.getString("SetupDialog.title.setBindingAddress"), //$NON-NLS-1$
										JOptionPane.ERROR_MESSAGE);

								return;
							}
						}
					}
					catch (UnknownHostException e) {
						JOptionPane.showMessageDialog(frame,
								messages.getString("SetupDialog.label.invalidNetworkInterface"), //$NON-NLS-1$
								messages.getString("SetupDialog.title.invalidNetworkInterface"), //$NON-NLS-1$
								JOptionPane.ERROR_MESSAGE);

						inetAddress = inputPrefs.getBindAddress();
					}

					String zeroConfImplementationStr = SyncDocs.JM_DNS_IMPLEMENTATION;
					if (zeroConfImplementation != null) {
						zeroConfImplementationStr =
								zeroConfImplementation.getSelectedItem().toString();
					}

					prefs = new Preferences(port, bonjourName, inetAddress,
							checkVersionFreqInDays, bindAll,
							useUserDirForDownloads.isSelected(),
							zeroConfImplementationStr);

					setVisible(false);
				}
				catch (NumberFormatException e) {
					// Leave port set to 0.
				}

				if (port < 1 || port > 65535) {
					JOptionPane.showMessageDialog(finalFrame, messages.getString("SetupDialog.label.invalidPort"), messages.getString("SetupDialog.title.invalidPort"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$ //$NON-NLS-2$
				}
			}
		};

		Action cancelActionListener = new AbstractAction() {
			private static final long serialVersionUID = -3865916887584095101L;

			public void actionPerformed(ActionEvent actionEvent) {
				setVisible(false);
			}
		};

		if (allowCancel) {
			InputMap inputMap = rootPane.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
		    inputMap.put(KeyStroke.getKeyStroke("ESCAPE"), "ESCAPE"); //$NON-NLS-1$ //$NON-NLS-2$
		    rootPane.getActionMap().put("ESCAPE", cancelActionListener); //$NON-NLS-1$
		}

		InputMap inputMap = rootPane.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
		inputMap.put(KeyStroke.getKeyStroke("ENTER"), "ENTER"); //$NON-NLS-1$ //$NON-NLS-2$
	    rootPane.getActionMap().put("ENTER", okActionListener); //$NON-NLS-1$

	    okButton.setAction(okActionListener);
		okButton.setText(messages.getString("SyncDocs.button.OK")); //$NON-NLS-1$

		if (allowCancel) {
			cancelButton.setAction(cancelActionListener);
			cancelButton.setText(messages.getString("SyncDocs.button.cancel")); //$NON-NLS-1$
		}

		if (System.getProperty("os.name").startsWith("Mac")) { //$NON-NLS-1$ //$NON-NLS-2$
			if (allowCancel) { buttonPanel.add(cancelButton); }
			buttonPanel.add(okButton);
		} else {
			buttonPanel.add(okButton);
			if (allowCancel) { buttonPanel.add(cancelButton); }
		}

		myPanel.add(BorderLayout.SOUTH, buttonPanel);

		if (zeroConfImplementation != null &&
				zeroConfImplementation.getSelectedIndex() == 0)
		{
			wirelessAddressButton.setEnabled(true);
			finalbindAddressLabel.setEnabled(true);
			changeBindAddressButton.setEnabled(true);
			bindAllButton.setEnabled(true);
		}
		else {
			wirelessAddressButton.setEnabled(false);
			finalbindAddressLabel.setEnabled(false);
			changeBindAddressButton.setEnabled(false);
			bindAllButton.setEnabled(false);
		}

		pack();

		// Get the width with the "Wireless Address" at the maximum size that
		// could be displayed (255.255.255.255), then update the text.
		// This code is necessary to avoid weird display issues on Windows
		// and Mac OS X where all fields are displayed truncated when the
		// "Wireless Address" line can't fit inside the dialog box.
		double width = wirelessAddressButton.getPreferredSize().getWidth();
		width += bindAddressLabel.getPreferredSize().getWidth();
		width += changeBindAddressButton.getPreferredSize().getWidth();
		int totalWidth = ((int) width) + (INSET_SIZE * 3);

		int height = (int) getMinimumSize().getHeight();
		setMinimumSize(new Dimension(totalWidth, height));
		bindAddressLabel.setText(bindAddress);

		setLocationRelativeTo(frame);
		setVisible(true);
	}

	public boolean wasOkSelected() {
		return okSelected;
	}

	public Preferences getPrefs() {
		return prefs;
	}

}
