/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.syncdocs;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.URL;
import java.net.URLConnection;
import java.net.UnknownHostException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Properties;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.InputMap;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.KeyStroke;

import ca.quine.web.IRequestHandler;
import ca.quine.web.ParsedHTTPRequest;
import ca.quine.web.TeenyWeb;

public class SyncDocs
		implements FileReceivedHandler, IRequestHandler,
		IExceptionCallback, ITerminatedCallback, PropertyChangeListener
{
	public static final String JM_DNS_IMPLEMENTATION = "JmDNS";
	public static final String BONJOUR_IMPLEMENTATION = "Bonjour";

	public static final String VERSION = "0.6.7"; //$NON-NLS-1$

	private static final String XML_DATE_FORMAT = "yyyy-MM-dd'T'HH:mm:ssZZZ"; //$NON-NLS-1$

	private static int BUFFER_SIZE = 4096;

	private static final long ONE_DAY_IN_MILLIS = 24 * 60 * 60 * 1000;

	private static final String FILE_TYPE = "filetype"; //$NON-NLS-1$
	private static final String APP_ID = "appId"; //$NON-NLS-1$
	private static final String APP_NAME = "appName"; //$NON-NLS-1$
	private static final String DIR = "dir"; //$NON-NLS-1$

	private static Pattern SYNCDOCS_VERSION_PATTERN = Pattern.compile("SyncDocs-([0-9\\.]*).zip"); //$NON-NLS-1$

	private static final int INSET = 20;

	private int port;
	private int recommendedPort;
	private boolean recommendedPortInUse;
	private String serviceName;
	private InetAddress bindAddress;
	private String zeroConfImplementation;
	private int checkVersionFreqInDays;
	private long lastCheckedVersion = 0;
	private boolean bindAll = true;
	private boolean useUserDirForDownloads = false;

	private TeenyWeb webServer;

	private DropPanel dropPanel;
	private JFrame appFrame;

	JPanel buttonPanel;
	JLabel welcomeLabel;
	JButton languagesButton;
	JButton showDownloadsButton;
	JButton setupButton;
	JButton syncButton;
	JButton removeButton;
	JButton helpButton;

	// Keep track of which files were in the list of files we returned so that
	// the application can remove them from the list.
	private ListHashMap fileMap;

	private static final HashMap mimeTypes = new HashMap();

	private List zeroConfThreads;

	private Messages messages;

	/* package */ static final List SHOW_DIR_APP_LIST;
	/* package */ static final List SHOW_SITE_APP_LIST;

	static {
		SHOW_DIR_APP_LIST = new ArrayList();
		SHOW_DIR_APP_LIST.add("nautilus");
		SHOW_DIR_APP_LIST.add("konqueror");

		SHOW_SITE_APP_LIST = new ArrayList();
		SHOW_SITE_APP_LIST.add("firefox");
		SHOW_SITE_APP_LIST.add("opera");
		SHOW_SITE_APP_LIST.add("konqueror");
		SHOW_SITE_APP_LIST.add("epiphany");
		SHOW_SITE_APP_LIST.add("mozilla");
		SHOW_SITE_APP_LIST.add("netscape");
	}

	public SyncDocs() {
		mimeTypes.put("gif", "image/gif"); //$NON-NLS-1$ //$NON-NLS-2$
		mimeTypes.put("jpg", "image/jpeg"); //$NON-NLS-1$ //$NON-NLS-2$
		mimeTypes.put("jpeg", "image/jpeg"); //$NON-NLS-1$ //$NON-NLS-2$
		mimeTypes.put("png", "image/png"); //$NON-NLS-1$ //$NON-NLS-2$
		mimeTypes.put("pdf", "application/pdf"); //$NON-NLS-1$ //$NON-NLS-2$

		fileMap = new ListHashMap();

		zeroConfThreads = new ArrayList();

		messages = new Messages();
		messages.registerListener(this);

		Runtime.getRuntime().addShutdownHook(new Thread() {
            public void run() {
                endApp();
            }
        });;
	}

	public static boolean isEmpty(String input) {
		if (input == null) { return true; }
		if (input.trim().length() == 0) { return true; }

		return false;
	}

	private void endApp() {
		// Just in case the app stops before it was fully started,
		// make sure that we have the appFrame.
		if (appFrame != null) {
			// Since the event shutdown hook is running, it might block.
			new Thread() {
				public void run() {
					appFrame.setVisible(false);
				}
			}.start();
		}

		if (port == 0 || isEmpty(serviceName) || dropPanel == null) {
			// Don't save anything if we haven't got everything set up
			// successfully.
		}
		else {
			File setupFile = getSetupFile();
			System.out.println("Saving preferences to '" + setupFile.getAbsolutePath() + "'."); //$NON-NLS-1$ //$NON-NLS-2$
			saveSetup(port, serviceName, dropPanel.getFilenameList());
			System.out.println("Preferences successfully saved."); //$NON-NLS-1$
		}

		System.out.flush();
		System.out.println("Shutting down web server and Bonjour."); //$NON-NLS-1$

		unbindZeroConf(5000);

		try {
			if (webServer != null) {
				webServer.stopServer();
			}

			System.out.println("Clean web server exit.");
		}
		catch (IOException e) {
			// If it fails, just bring it down hard.
			System.out.println("Killing the web server.");
			System.out.flush();

			System.exit(0);
		}
	}

	private void unbindZeroConf(int timeUntilExit) {
		List prevZeroConfThreads = zeroConfThreads;

		String threads = oneOrMoreWithPrefix("There is ", "There are ",
				zeroConfThreads.size(), " thread", " threads");

		System.out.println(threads + " to stop.");

		Iterator zeroConfThreadIter = zeroConfThreads.iterator();
		while (zeroConfThreadIter.hasNext()) {
			ZeroConfThread thread = (ZeroConfThread) zeroConfThreadIter.next();
			System.out.println("Calling unbind on " + thread.getName() + ".");
			thread.unbind();
		}

		zeroConfThreads = new ArrayList();

		if (timeUntilExit > 0) {
			long exitTime = System.currentTimeMillis() + timeUntilExit;

			synchronized (prevZeroConfThreads) {
				while (true) {
					if (prevZeroConfThreads.size() == 0) {
						return;
					}

					long now = System.currentTimeMillis();
					if (now >= exitTime) {
						System.err.println("Bonjour is taking too long to shutdown.  Force quitting.");
						System.exit(0);
					}

					try {
						prevZeroConfThreads.wait(exitTime - now);
					}
					catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}
		}
	}

	public static void main(String[] args) {
		SyncDocs syncDocs = new SyncDocs();
		syncDocs.run();
	}

	private void run() {
		System.out.print("SyncDocs version " + VERSION + " starting, running on:  " + System.getProperty("os.name"));
		System.out.print(" (OS version " + System.getProperty("os.version") + ")");
		System.out.println(" with Java " + System.getProperty("java.version") + ".");

		final JFrame f = new JFrame("SyncDocs"); //$NON-NLS-1$
		appFrame = f;
		f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		f.getContentPane().setLayout(new BorderLayout());

		Action locationResetListener = new AbstractAction() {
			private static final long serialVersionUID = 1L;

			public void actionPerformed(ActionEvent actionEvent) {
				f.setLocation(0, 0);
			}
		};

		InputMap inputMap = f.getRootPane().getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
		KeyStroke ctrlR = KeyStroke.getKeyStroke(KeyEvent.VK_R, Event.CTRL_MASK);
	    inputMap.put(ctrlR, "CTRL-R"); //$NON-NLS-1$
	    f.getRootPane().getActionMap().put("CTRL-R", locationResetListener); //$NON-NLS-1$

		Action sizeResetListener = new AbstractAction() {
			private static final long serialVersionUID = 1L;

			public void actionPerformed(ActionEvent actionEvent) {
				f.setSize(600, 300);
				resizeToButtonPanel();
			}
		};

		KeyStroke ctrlS = KeyStroke.getKeyStroke(KeyEvent.VK_S, Event.CTRL_MASK);
	    inputMap.put(ctrlS, "CTRL-S"); //$NON-NLS-1$
	    f.getRootPane().getActionMap().put("CTRL-S", sizeResetListener); //$NON-NLS-1$

	    welcomeLabel = new JLabel(messages.getString("SyncDocs.label.dragFilesBelow")); //$NON-NLS-1$
		f.getContentPane().add(welcomeLabel, BorderLayout.NORTH);

		dropPanel = new DropPanel(this);
		f.getContentPane().add(dropPanel, BorderLayout.CENTER);

		createButtonPanel(f, dropPanel);
		f.getContentPane().add(buttonPanel, BorderLayout.SOUTH);

		f.pack();
		f.setVisible(true);

		if (!loadSetup(f)) {
			boolean allowCancel = false;
			setRecommendedPort();
			port = recommendedPort;
			runSetup(f, this, allowCancel);
		}
	}

	private File getSetupFile() {
		String setupFilename;

		if (isRunningOnWindows()) {
			setupFilename = "SyncDocs.conf"; //$NON-NLS-1$
		}
		else {
			setupFilename = ".syncdocs"; //$NON-NLS-1$
		}

		return new File(System.getProperty("user.home") + File.separator + setupFilename); //$NON-NLS-1$
	}

	private boolean isRunningOnWindows() {
		return System.getProperty("os.name").indexOf("Windows") > -1;
	}

	private boolean loadSetup(JFrame f) {
		setRecommendedPort();

		try {
			File setupFile = getSetupFile();
			Properties props = new Properties();

			if (setupFile.canRead()) {
				InputStream is = new FileInputStream(setupFile);
				props.load(is);
				is.close();
			}
			else {
				port = recommendedPort;
				serviceName = "SyncDocs"; //$NON-NLS-1$
			}

			// Read the files first, so that if there's a problem with
			// other properties, the files are still handled and will
			// show up correctly.
		    List files = readFiles(props, "filename"); //$NON-NLS-1$
		    dropPanel.addFiles(files);

		    try {
		    	port = Integer.parseInt(props.getProperty("port")); //$NON-NLS-1$
		    }
		    catch (Exception e) {
		    	port = recommendedPort;
		    }

		    if (serviceName == null) {
		    	serviceName = props.getProperty("serviceName"); //$NON-NLS-1$
		    }

		    String bindAddressString = props.getProperty("bindAddress"); //$NON-NLS-1$

		    String bindAllProp = props.getProperty("bind.all"); //$NON-NLS-1$
		    if (bindAllProp == null) {
		    	List allInetAddresses = getAllInetAddresses();
		    	if (allInetAddresses.size() < 4) {
		    		bindAll = true;
		    	}
		    	else if (bindAddressString == null) {
		    		List ifaceList = NetworkInterfaceDialog.getAllNetworkInterfaces();
				    List networkInterfaces = NetworkInterfaceDialog.buildInterfaceList(ifaceList);

					bindAddress = SyncDocs.getUserChoiceOfInterface(messages,
							f, networkInterfaces, false);
		    	}
		    }
		    else {
		    	bindAll = parseBoolean(bindAllProp);
		    }

		    if (serviceName != null && serviceName.indexOf(".") > -1) { //$NON-NLS-1$
				JOptionPane.showMessageDialog(f,
						messages.getString("SyncDocs.error.bonjourNameHasPeriod"), //$NON-NLS-1$
						messages.getString("SyncDocs.title.invalidBonjourName"), //$NON-NLS-1$
						JOptionPane.ERROR_MESSAGE);
		    }

		    boolean isBonjourAvailable = isBonjourAvailable();

		    zeroConfImplementation = props.getProperty("zeroConf.impl");
		    if (zeroConfImplementation == null) {
		    	if (isBonjourAvailable) {
			    	System.out.println("SyncDocs now supports Apple's Bonjour implementation, and it's available.");
			    	System.out.println("Bonjour will be used instead of JmDNS.");
			    	zeroConfImplementation = BONJOUR_IMPLEMENTATION;
		    	}
		    	else {
			    	System.out.println("Bonjour is not available.  JmDNS will be used instead of Bonjour.");
			    	zeroConfImplementation = JM_DNS_IMPLEMENTATION;
		    	}
		    }
		    else {
		    	// A Zeroconf implementation has been configured.
			    if (BONJOUR_IMPLEMENTATION.equals(zeroConfImplementation)) {
			    	if (isBonjourAvailable) {
				    	System.out.println("Bonjour is available and has been configured in the setup.");
				    	System.out.println("Bonjour will be used.");
			    	}
			    	else {
			    		System.out.println("Bonjour was configured, but it's not available.");
				    	System.out.println("JmDNS will be used instead of Bonjour.");
				    	zeroConfImplementation = JM_DNS_IMPLEMENTATION;
			    	}
			    }
			    else {
			    	if (isBonjourAvailable) {
				    	System.out.println("Bonjour is available, but JmDNS has been configured in the setup.");
			    	}
			    	else {
			    		System.out.println("JmDNS was configured and Bonjour is not available.");
			    	}

			    	System.out.println("JmDNS will be used instead of Bonjour.");
			    	zeroConfImplementation = JM_DNS_IMPLEMENTATION;
			    }
		    }

		    try {
			    int width = Integer.parseInt(props.getProperty("frame.width")); //$NON-NLS-1$
			    int height = Integer.parseInt(props.getProperty("frame.height")); //$NON-NLS-1$

			    appFrame.setSize(width, height);
		    }
		    catch (Exception e) {
		    	// Ignore any errors if the size could not be restored.
		    }

		    try {
			    int x = Integer.parseInt(props.getProperty("frame.x")); //$NON-NLS-1$
			    int y = Integer.parseInt(props.getProperty("frame.y")); //$NON-NLS-1$

			    appFrame.setLocation(x, y);
		    }
		    catch (Exception e) {
		    	// Ignore any errors if the location could not be restored.
		    }

		    try {
			    checkVersionFreqInDays =
			    		Integer.parseInt(props.getProperty("check.frequency")); //$NON-NLS-1$

			    lastCheckedVersion =
		    			Long.parseLong(props.getProperty("check.last")); //$NON-NLS-1$
		    }
		    catch (Exception e) {
		    	// Ignore any errors if the frequency to check for updates
		    	// could not be restored.
		    }

		    if (checkVersionFreqInDays != 0) {
		    	long nextCheck = lastCheckedVersion +
		    			(checkVersionFreqInDays * ONE_DAY_IN_MILLIS);

		    	if (System.currentTimeMillis() > nextCheck) {
		    		isNewVersionAvailable();
		    	}
		    	else {
		    		System.out.println("No need to check for a new version until:  " + new Date(nextCheck) + "."); //$NON-NLS-1$ //$NON-NLS-2$
		    	}
		    }

		    String useUserDir = props.getProperty("useUserDirForDownloads");
		    if ("true".equals(useUserDir)) {
		    	useUserDirForDownloads = true;
		    }

		    if (bindAddressString != null) {
		    	try {
		    		bindAddress = NetworkInterfaceDialog.convertToInetAddress(bindAddressString);
		    	}
		    	catch (UnknownHostException e) {
					JOptionPane.showMessageDialog(f,
							messages.getString("SyncDocs.error.invalidNetworkInterface"), //$NON-NLS-1$
							messages.getString("SyncDocs.title.invalidNetworkInterface"), //$NON-NLS-1$
							JOptionPane.ERROR_MESSAGE);

					bindAddress = null;
					return false;
		    	}
		    }

		    if (!bindWebServer(f)) {
		    	return false;
		    }

			bindZeroConf(f);

			return true;
		}
		catch (IOException e) {
			// Fall through and return false.
		}

		return false;
	}

	/**
	 * This method was added for those who don't have JDK 1.5.
	 * TODO:  It can be removed if Java 1.4 support is no longer
	 * required.
	 * 
	 * @param input
	 * @return
	 */
	private boolean parseBoolean(String input) {
		return "true".equalsIgnoreCase(input);
	}

	private boolean bindWebServer(JFrame f) {
	    // Try for 2 seconds to find an open port.
	    long startTime = System.currentTimeMillis();

	    System.out.println("Attempting to bind on port " + port + ".");

		do {
		    try {
				webServer = new TeenyWeb(port, this);
				System.out.println("Web server ready, listening on port " + port + "."); //$NON-NLS-1$ //$NON-NLS-2$
				break;
			}
			catch (IOException e) {
				if (port == recommendedPort) {
					recommendedPortInUse = true;
				}

				port++;
				if (port > 65535) { port = 8080; }
				if (startTime + 2000 < System.currentTimeMillis()) {
					System.out.println("Unable to find a port after checking for 2 seconds.");
					handleBindException(f);
					webServer = null;
					return false;
				}
			}
		}
		while (true);

		return true;
	}

	public static boolean isBonjourAvailable() {
		try {
			if (BonjourThread.isAvailable()) {
				return true;
			}

			return false;
		}
		catch (NoClassDefFoundError e) {
			return false;
		}
		catch (UnsatisfiedLinkError e) {
			System.out.println("Bonjour is not available due to an UnsatisfiedLinkError:  " + e.getMessage());

			return false;
		}
	}

	private void bindZeroConf(JFrame f) {
		if (zeroConfImplementation.equals(BONJOUR_IMPLEMENTATION)) {
			BonjourThread thread = new BonjourThread(bindAddress, serviceName, port, zeroConfThreads, this, this);
			zeroConfThreads.add(thread);
			thread.start();

			return;
		}

		if (!bindAll) {
			JmDNSThread thread = new JmDNSThread(bindAddress, serviceName, port, zeroConfThreads, this, this);
			zeroConfThreads.add(thread);
			thread.start();
		}
		else {
			System.out.println("Binding to all addresses:  " + System.currentTimeMillis()); //$NON-NLS-1$

			List ifaceList;
			try {
				ifaceList = NetworkInterfaceDialog.getAllNetworkInterfaces();
			}
			catch (SocketException e) {
				JOptionPane.showMessageDialog(f,
						messages.getString("SyncDocs.error.unableToFindInterfaces"), //$NON-NLS-1$
						messages.getString("SyncDocs.title.unableToFindInterfaces"), //$NON-NLS-1$
						JOptionPane.ERROR_MESSAGE);

				return;
			}

			Iterator ifaceIter = ifaceList.iterator();
			while (ifaceIter.hasNext()) {
				NetworkInterface iface = (NetworkInterface) ifaceIter.next();

				Enumeration interfaceEnum = iface.getInetAddresses();
			    while (interfaceEnum.hasMoreElements()) {
			    	InetAddress address = (InetAddress) interfaceEnum.nextElement();

			    	if (address.isLoopbackAddress()) {
			    		continue;
			    	}

			    	// Reduce it down to just IPv4 addresses to cut down on startup time.
			    	// The iPhone client doesn't support IPv6 anyway.
			    	if (address.getAddress().length == 4) {
						JmDNSThread thread = new JmDNSThread(address, serviceName, port, zeroConfThreads, this, this);
						zeroConfThreads.add(thread);
						thread.start();
			    	}
//			    	else {
//				    	System.out.println("Found IPv6 local address " + address);
//			    	}
				}
			}

			System.out.println("All threads started:  " + System.currentTimeMillis()); //$NON-NLS-1$
		}
	}

	public static List getAllInetAddresses() throws SocketException {
		List retVal = new ArrayList();

		List ifaceList = NetworkInterfaceDialog.getAllNetworkInterfaces();
		Iterator ifaceIter = ifaceList.iterator();
		while (ifaceIter.hasNext()) {
			NetworkInterface iface = (NetworkInterface) ifaceIter.next();

			Enumeration interfaceEnum = iface.getInetAddresses();
		    while (interfaceEnum.hasMoreElements()) {
		    	InetAddress address = (InetAddress) interfaceEnum.nextElement();

		    	if (address.isLoopbackAddress()) {
		    		continue;
		    	}

		    	// Reduce it down to just IPv4 addresses to cut down on startup time.
		    	// The iPhone client doesn't support IPv6 yet anyway.
		    	if (address.getAddress().length == 4) {
					retVal.add(address.getAddress());
		    	}
			}
		}

		return retVal;
	}

	public Boolean isNewVersionAvailable() {
		String syncDocsUrl = "http://sourceforge.net/projects/syncdocs/files/"; //$NON-NLS-1$
		String notesUrl = null;
		BufferedReader br = null;
		boolean checkFailed = true;

		try {
			System.out.println("Checking for latest version at:  '" + syncDocsUrl + "'."); //$NON-NLS-1$ //$NON-NLS-2$
			URL url = new URL(syncDocsUrl);

			URLConnection conn = url.openConnection();
			br = new BufferedReader(new InputStreamReader(conn.getInputStream()));

			Version currentVersion = new Version(VERSION);

			Pattern notesPattern = null;
			boolean newVersion = false;
			String line = null;
			while ((line = br.readLine()) != null) {
				Matcher verMatcher = SYNCDOCS_VERSION_PATTERN.matcher(line);
				Matcher notesMatcher;
				if (notesPattern != null) {
					notesMatcher = notesPattern.matcher(line);
				}
				else {
					notesMatcher = null;
				}

				if (verMatcher.find()) {
		    		lastCheckedVersion = System.currentTimeMillis();

		    		String versionToCheckAgainst = verMatcher.group(1);
		    		if (!currentVersion.isVersionAtLeast(versionToCheckAgainst)) {
						newVersion = true;
						System.out.println("Found new version:  '" + verMatcher.group(1) + "'.");
						if (notesPattern == null) {
							notesPattern = Pattern.compile("\\/project\\/syncdocs\\/syncdocs\\/" + verMatcher.group(1) + "\\/CHANGELOG.txt");
						}
					}
				}

				if (notesMatcher != null && notesMatcher.find() && notesUrl == null) {
					notesUrl = "http://downloads.sourceforge.net" + notesMatcher.group(0).replaceAll("&amp;", "&"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
				}
			}

			if (newVersion) {
				Object[] options;
				if (notesUrl != null) {
					options = new Object[3];
					options[0] = messages.getString("SyncDocs.button.download"); //$NON-NLS-1$
					options[1] = messages.getString("SyncDocs.button.whatsNew"); //$NON-NLS-1$
					options[2] = messages.getString("SyncDocs.button.later"); //$NON-NLS-1$
				}
				else {
					options = new Object[2];
					options[0] = messages.getString("SyncDocs.button.download"); //$NON-NLS-1$
					options[1] = messages.getString("SyncDocs.button.later"); //$NON-NLS-1$
				}

				do {
					int result = JOptionPane.showOptionDialog(null,
							messages.getString("SyncDocs.label.newVersionAvailable"), //$NON-NLS-1$
							messages.getString("SyncDocs.title.newVersionAvailable"), //$NON-NLS-1$
							JOptionPane.DEFAULT_OPTION,
							JOptionPane.INFORMATION_MESSAGE,
							null, options, options[0]);

					if (result == 0) {
						String downloadUrl = "http://sourceforge.net/project/platformdownload.php?group_id=235152"; //$NON-NLS-1$

						SyncDocs.launch(downloadUrl, SHOW_SITE_APP_LIST,
								messages.getString("SyncDocs.error.unableToLaunchBrowser"),
								messages.getString("SyncDocs.error.unableToLaunchBrowser2"));
					}

					if (options.length == 3 && result == 1) {
						SyncDocs.launch(notesUrl, SHOW_SITE_APP_LIST,
								messages.getString("SyncDocs.error.unableToLaunchBrowser"),
								messages.getString("SyncDocs.error.unableToLaunchBrowser2"));

						continue;
					}

					break;
				}
				while (true);

				checkFailed = false;
				return Boolean.TRUE;
			}

			checkFailed = false;
		}
		catch (Exception e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(null, messages.getString("SyncDocs.error.failedTryingToCheck")); //$NON-NLS-1$
		}
		finally {
			if (br != null) {
				try {
					br.close();
				}
				catch (IOException e) { /* Ignore it */ }
			}
		}

		if (checkFailed) { return null; }

		return Boolean.FALSE;
	}

	public static InetAddress getNetworkInterface(Messages messages, JFrame f)
			throws SocketException
	{
		InetAddress networkInterface = null;

		List ifaceList = NetworkInterfaceDialog.getAllNetworkInterfaces();

		// If there's more than the loopback and another interface, then
		// ask the user which one is the wireless.
	    List networkInterfaces = NetworkInterfaceDialog.buildInterfaceList(ifaceList);

	    if (ifaceList.size() > 2) {
			networkInterface = getUserChoiceOfInterface(messages, f,
					networkInterfaces, false);
		}
	    else {
	    	// If the non-loopback interface has a single IP, then use it
	    	// and keep it for later.
	    	if (networkInterfaces.size() == 1) {
	    		NetworkDetails details = (NetworkDetails) networkInterfaces.get(0);
	    		return details.getInetAddress();
	    	}
	    }

		return networkInterface;
	}

	public static InetAddress getUserChoiceOfInterface(Messages messages,
			JFrame f, List networkInterfaces, boolean allowCancel)
	{
		NetworkInterfaceDialog dialog = new NetworkInterfaceDialog(f,
				messages,
				messages.getString("SyncDocs.label.pleaseSelectWirelessInterface"), //$NON-NLS-1$
				allowCancel, networkInterfaces);

		return dialog.getSelectedAddress();
	}

	private List readFiles(Properties props, String prefix)
	{
		ArrayList retVal = new ArrayList();

		int i = 0;
		while (true) {
			String filename = props.getProperty(prefix + "." + i); //$NON-NLS-1$

			if (filename != null) {
				String appId = props.getProperty(prefix + "." + i + ".appId"); //$NON-NLS-1$ //$NON-NLS-2$

				if (isEmpty(appId)) {
					appId = FileEntry.SYNC_WITH_NEXT;
				}

				retVal.add(new FileEntry(filename, appId));
			}
			else {
				break;
			}

			i++;
		}

		return retVal;
	}

	private void saveSetup(int port, String serviceName, List filenameList) {
		File setupFile = getSetupFile();

		try {
			OutputStream os = new FileOutputStream(setupFile);

			Properties props = new Properties();
			props.setProperty("port", port + ""); //$NON-NLS-1$ //$NON-NLS-2$
			props.setProperty("serviceName", serviceName); //$NON-NLS-1$

			Iterator filenameIter = filenameList.iterator();
			int i = 0;
			while (filenameIter.hasNext()) {
				FileEntry fileEntry = (FileEntry) filenameIter.next();

				props.setProperty("filename." + i, fileEntry.getFileName()); //$NON-NLS-1$
				props.setProperty("filename." + i + ".appId", fileEntry.getAppId()); //$NON-NLS-1$ //$NON-NLS-2$
				i++;
			}

			props.setProperty("frame.width", ((int) appFrame.getSize().getWidth()) + ""); //$NON-NLS-1$ //$NON-NLS-2$
			props.setProperty("frame.height", ((int) appFrame.getSize().getHeight()) + ""); //$NON-NLS-1$ //$NON-NLS-2$
			props.setProperty("frame.x", appFrame.getX() + ""); //$NON-NLS-1$ //$NON-NLS-2$
			props.setProperty("frame.y", appFrame.getY() + ""); //$NON-NLS-1$ //$NON-NLS-2$

			StringBuffer sb = new StringBuffer();
			if (bindAddress != null) {
				NetworkInterfaceDialog.convertToDotted(sb, bindAddress.getAddress());
				props.setProperty("bindAddress", sb.toString()); //$NON-NLS-1$
			}

			props.setProperty("bind.all", bindAll + ""); //$NON-NLS-1$ //$NON-NLS-2$

			props.setProperty("zeroConf.impl", zeroConfImplementation);

			props.setProperty("check.frequency", checkVersionFreqInDays + ""); //$NON-NLS-1$ //$NON-NLS-2$
			props.setProperty("check.last", lastCheckedVersion + ""); //$NON-NLS-1$ //$NON-NLS-2$

			props.setProperty("useUserDirForDownloads", useUserDirForDownloads + "");

			props.store(os, "Configuration file for SyncDocs"); //$NON-NLS-1$
			os.flush();
			os.close();
		}
		catch (IOException e) {
			JOptionPane.showMessageDialog(null, messages.getString("SyncDocs.error.unableToSaveConfigFile") + setupFile.getAbsolutePath() + messages.getString("SyncDocs.error.unableToSaveConfigFile2")); //$NON-NLS-1$ //$NON-NLS-2$
		}
	}

	private void runSetup(JFrame f, IRequestHandler handler, boolean allowCancel) {
		setRecommendedPort();

		int usePort = recommendedPort;
		if (recommendedPortInUse) {
			usePort = 0;
		}

		do {
			if (!bindAll && bindAddress == null) {
				try {
					bindAddress = getNetworkInterface(messages, f);
				}
				catch (SocketException e) {
					JOptionPane.showMessageDialog(f,
							messages.getString("SyncDocs.error.noNetworkConnection"), //$NON-NLS-1$
							messages.getString("SyncDocs.title.noNetworkConnection"), //$NON-NLS-1$
							JOptionPane.ERROR_MESSAGE);

					continue;
				}
			}

			Preferences prefs = new Preferences(port, serviceName, bindAddress,
					checkVersionFreqInDays, bindAll, useUserDirForDownloads,
					zeroConfImplementation);

			SetupDialog setupDialog =
					new SetupDialog(this, f, messages.getString("SyncDocs.title.setup"), allowCancel, usePort, prefs); //$NON-NLS-1$

			if (!allowCancel) {
				// If we're not allowing cancel, then we're setting everything up.
				// Make sure the port is 0 so that any port will be treated as a
				// "change" in the code below, and we'll set up the web server.
				prefs.setPort(0);
			}

			if (allowCancel && !setupDialog.wasOkSelected()) {
				break;
			}

			if (!allowCancel || setupDialog.wasOkSelected()) {
				Preferences newPrefs = setupDialog.getPrefs();

				zeroConfImplementation = newPrefs.getZeroConfImplementation();

				useUserDirForDownloads = newPrefs.isUseUserDirForDownloads();
				checkVersionFreqInDays = newPrefs.getCheckVersionFreqInDays();

				if (newPrefs.isBindAll() != prefs.isBindAll() ||
						!safeEquals(newPrefs.getBindAddress(), prefs.getBindAddress()))
				{
					unbindZeroConf(-1);

					bindAll = newPrefs.isBindAll();
					bindAddress = newPrefs.getBindAddress();
				}

				if ((newPrefs.getPort() != prefs.getPort()) || webServer == null) {
					if (webServer != null) {
						try {
							webServer.stopServer();
						}
						catch (IOException e) {
							JOptionPane.showMessageDialog(f,
									messages.getString("SyncDocs.error.cannotStopServer"), //$NON-NLS-1$
									messages.getString("SyncDocs.title.cannotStopServer"), //$NON-NLS-1$
									JOptionPane.ERROR_MESSAGE);
						}

						webServer = null;
					}

					port = newPrefs.getPort();
					if (!bindWebServer(f)) {
						continue;
					}

					serviceName = newPrefs.getBonjourName();
					bindAddress = newPrefs.getBindAddress();
					bindAll = newPrefs.isBindAll();

				    if (webServer == null) {
				    	continue;
				    }
				}

				serviceName = newPrefs.getBonjourName();
				bindAddress = newPrefs.getBindAddress();
				bindAll = newPrefs.isBindAll();

			    bindZeroConf(f);
		    	break;
			}

			// If we got here the user clicked "OK" but didn't change anything.
			// We wouldn't have dismissed the dialog if there were problems
			// with the input data, so just drop out of the loop.
			break;
		}
		while (true);
	}

	private boolean safeEquals(InetAddress address1, InetAddress address2) {
		if (address1 == null && address2 == null) { return true; }
		if (address1 == null && address2 != null) { return false; }

		// "address1" can't be null here.
		return address1.equals(address2);
	}

	private void handleBindException(JFrame f) {
		JOptionPane.showMessageDialog(f,
				messages.getString("SyncDocs.error.unableToFindOpenPort"), //$NON-NLS-1$
				messages.getString("SyncDocs.title.unableToFindOpenPort"), //$NON-NLS-1$
				JOptionPane.ERROR_MESSAGE);

		if (port == recommendedPort) {
			recommendedPortInUse = true;
		}
	}

	private void setRecommendedPort() {
		if (recommendedPort != 0) { return; }

		if (isRunningOnWindows()) {
			recommendedPort = 80;
		}
		else {
			recommendedPort = 8080;
		}
	}

	private void createButtonPanel(final JFrame f,
			final DropPanel dropPanel)
	{
		final SyncDocs thisSyncDocs = this;

		buttonPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));

		languagesButton = new JButton();
		languagesButton.setAction(new AbstractAction() {
			private static final long serialVersionUID = 8956849127624533646L;

			public void actionPerformed(ActionEvent arg0) {
				Locale[] availableLocales = messages.getAvailableLocales();
				Locale currentLocale = messages.getCurrentLocale();

				Object[] possibleValues = new String[availableLocales.length];

				int j = 0;
				for (int i = 0; i < possibleValues.length; i++) {
					// For some reason, Portuguese is displaying in all lower
					// case.
					possibleValues[i] = initialCap(
							availableLocales[i].getDisplayLanguage(
							availableLocales[i]));

					if (availableLocales[i].equals(currentLocale)) {
						j = i;
					}
				}

				Object selectedKey = JOptionPane.showInputDialog(null,
						messages.getString("SyncDocs.label.selectALanguage"), //$NON-NLS-1$
						messages.getString("SyncDocs.title.selectALanguage"), //$NON-NLS-1$
						JOptionPane.INFORMATION_MESSAGE, null,
						possibleValues, possibleValues[j]);

				if (selectedKey != null) {
					Locale selectedLocale = (Locale) getSelectedValue(selectedKey,
							possibleValues, availableLocales);

					messages.setLocale(selectedLocale);
				}
			}

			private Object initialCap(String input) {
				return Character.toUpperCase(input.charAt(0))
						+ input.substring(1);
			}

			private Object getSelectedValue(Object selectedKey, Object[] keys, Object[] values) {
				for (int i = 0; i < keys.length; i++) {
					Object object = keys[i];
					if (selectedKey.equals(object)) { 
						return values[i];
					}
				}

				return null;
			}
		});
		languagesButton.setText(messages.getString("SyncDocs.button.languages")); //$NON-NLS-1$
		buttonPanel.add(languagesButton);

		showDownloadsButton = new JButton();
		showDownloadsButton.setAction(new AbstractAction() {
			private static final long serialVersionUID = 8956849127624533646L;

			public void actionPerformed(ActionEvent arg0) {
				File file = getDownloadsDirectory(null);
				if (!file.exists()) {
					if (!file.mkdirs()) {
						JOptionPane.showMessageDialog(null, messages.getString("SyncDocs.error.unableToCreateDirectory") + file + messages.getString("SyncDocs.error.unableToCreateDirectory2")); //$NON-NLS-1$ //$NON-NLS-2$
						return;
					}
				}

				try {
					SyncDocs.launch(file.getCanonicalPath(), SHOW_DIR_APP_LIST,
							thisSyncDocs.getLocalisedString("SyncDocs.error.unableToShowDirectory"),
							thisSyncDocs.getLocalisedString("SyncDocs.error.unableToShowDirectory2"));
				}
				catch (IOException e) {
					JOptionPane.showMessageDialog(null, messages.getString("SyncDocs.error.unableToShowDirectory") + file + messages.getString("SyncDocs.error.unableToShowDirectory2")); //$NON-NLS-1$ //$NON-NLS-2$
					e.printStackTrace();
				}
			}
		});
		showDownloadsButton.setText(messages.getString("SyncDocs.button.showDownloads")); //$NON-NLS-1$
		buttonPanel.add(showDownloadsButton);

		setupButton = new JButton();
		setupButton.setAction(new AbstractAction() {
			private static final long serialVersionUID = 8956849127624533646L;

			public void actionPerformed(ActionEvent arg0) {
				runSetup(f, thisSyncDocs, true);
			}

		});
		setupButton.setText(messages.getString("SyncDocs.button.setup")); //$NON-NLS-1$
		buttonPanel.add(setupButton);

		syncButton = new JButton();
		syncButton.setAction(new AbstractAction() {
			private static final long serialVersionUID = -6881837535566533056L;

			public void actionPerformed(ActionEvent arg0) {
				JOptionPane.showMessageDialog(null, messages.getString("SyncDocs.label.launchTheApp")); //$NON-NLS-1$
			}

		});
		syncButton.setText(messages.getString("SyncDocs.button.sync")); //$NON-NLS-1$

		buttonPanel.add(syncButton);

		removeButton = new JButton();
		AbstractAction removeActionListener = new AbstractAction() {
			private static final long serialVersionUID = 7736392060870073668L;

			public void actionPerformed(ActionEvent arg0) {
				dropPanel.removeSelected();
			}
		};
		removeButton.setAction(removeActionListener);
		removeButton.setText(messages.getString("SyncDocs.button.remove")); //$NON-NLS-1$

		InputMap inputMap = f.getRootPane().getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
	    inputMap.put(KeyStroke.getKeyStroke("DELETE"), "DELETE"); //$NON-NLS-1$ //$NON-NLS-2$
	    f.getRootPane().getActionMap().put("DELETE", removeActionListener); //$NON-NLS-1$

		buttonPanel.add(removeButton);

		helpButton = new JButton();
		helpButton.setAction(new AbstractAction() {
			private static final long serialVersionUID = -8349564057549681219L;

			public void actionPerformed(ActionEvent arg0) {
				String url = "http://syncdocs.sf.net/getting-started.php"; //$NON-NLS-1$

				SyncDocs.launch(url, SHOW_SITE_APP_LIST,
						messages.getString("SyncDocs.error.unableToLaunchBrowser"),
						messages.getString("SyncDocs.error.unableToLaunchBrowser2"));
			}

		});
		helpButton.setText(messages.getString("SyncDocs.button.help")); //$NON-NLS-1$

		buttonPanel.add(helpButton);
	}

	public static void launch(String target, List appList,
			String preMessage, String postMessage)
	{
		try {
			String osName = System.getProperty("os.name"); //$NON-NLS-1$
			Process process = null;
			if (osName.startsWith("Windows")) { //$NON-NLS-1$
				process = Runtime.getRuntime().exec("rundll32 url.dll,FileProtocolHandler " + target); //$NON-NLS-1$
			}
			else if (osName.startsWith("Mac OS X")) {
				process = Runtime.getRuntime().exec("/usr/bin/open " + target); //$NON-NLS-1$
			}
			else {
				launchUnix(target, appList, preMessage, postMessage);

				return;
			}

			final Process finalProcess = process;
			Thread t = new Thread(new Runnable() {
				public void run() {
					try {
						finalProcess.waitFor();
						// Calling waitFor ensures we won't have any zombies
						// running around.  We don't need the return code.
					}
					catch (InterruptedException e) {
						// If the launch was interrupted, chances are the user
						// initiated the interruption, most likely by closing
						// the program.
					}
				}
			});
			t.start();
		}
		catch (IOException e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(null, preMessage + target + postMessage);
		}
	}

	private static void launchUnix(String target, List appList,
			String preMessage, String postMessage) throws IOException
	{
		try {
			String foundApp = null;
			Iterator appNameIter = appList.iterator();
			while (appNameIter.hasNext()) {
				String appName = (String) appNameIter.next();
				if (Runtime.getRuntime().exec(
						new String[] {"which", appName}).waitFor() == 0)
				{
					foundApp = appName;
					break;
				}
			}

			if (foundApp == null) {
				System.out.println("Unable to find any of:  " + appList + ".");
				JOptionPane.showMessageDialog(null,
						preMessage + target + postMessage);
			}
			else {
				Runtime.getRuntime().exec(new String[] {foundApp, target});
			}
		}
		catch (InterruptedException e) {
			// If the launch was interrupted, chances are the user initiated
			// the interruption, most likely by closing the program.
		}
	}

	public void receiveFileName(String filename) {
		// TODO:  Implement, if needed.
	}

	public boolean canRead(ParsedHTTPRequest request) {
		if (isBuiltin(request)) {
			return true;
		}

		String dir = request.getParameter(DIR);
		if (dir != null) {
			return checkInDirectory(request);
		}
		else {
			return (getFileInDownloadList(request) != null);
		}
	}

	private boolean checkInDirectory(ParsedHTTPRequest request) {
		File file = validateRequestDir(request);
		if (file == null) { return false; }

		return file.exists();
	}

	private File getFileInDownloadList(ParsedHTTPRequest request) {
		String requestedFile = request.getFilename();

		List filenameList = dropPanel.getFilenameList();
		Iterator filenameIter = filenameList.iterator();
		while (filenameIter.hasNext()) {
			FileEntry fileEntry = (FileEntry) filenameIter.next();

			String appId = request.getParameter("appId"); //$NON-NLS-1$
			if (!fileEntry.validateAppId(appId)) {
				continue;
			}

			File file = new File(fileEntry.getFileName());
			if (requestedFile.equals("/" + file.getName())) { //$NON-NLS-1$
				if (file.exists()) {
					return file;
				}
			}
		}

		return null;
	}

	private String handleBuiltins(ParsedHTTPRequest request) {
		String requestedFile = request.getFilename();

		boolean listButNoParams = false;
		if (requestedFile.equals("/list") || requestedFile.equals("/listWithDirs")) { //$NON-NLS-1$ //$NON-NLS-2$
			if (!request.hasParameters()) { //$NON-NLS-1$
				listButNoParams = true;
				System.out.println("List requested, but no parameters were supplied:  sending default page."); //$NON-NLS-1$
				return getIndexPage();
			}
		}

		if (requestedFile.equals("/hash")) { //$NON-NLS-1$
			if (request.getParameter("file") == null) { //$NON-NLS-1$
				System.out.println("Hash requested, but no file was supplied:  sending default page."); //$NON-NLS-1$
				return getIndexPage();
			}
		}

		boolean removeButNoParams = false;
		if (requestedFile.equals("/remove") && !request.hasParameters()) { //$NON-NLS-1$
			removeButNoParams = true;
			System.out.println("Remove requested, but no parameters were supplied:  sending default page."); //$NON-NLS-1$
			return getIndexPage();
		}

		if (requestedFile.equals("/") || //$NON-NLS-1$
				requestedFile.equals("/index.html") || //$NON-NLS-1$
				listButNoParams || removeButNoParams)
		{
			return getIndexPage();
		}

		String appId = request.getParameter(APP_ID);
		String filetypes = request.getParameter(FILE_TYPE);
		if (requestedFile.equals("/list") || requestedFile.equals("/listWithDirs")) { //$NON-NLS-1$
			if (filetypes != null || appId != null) {
				return getListPage(request, filetypes);
			}
			else {
				return getIndexPage();
			}
		}

		if (requestedFile.equals("/remove")) { //$NON-NLS-1$
			if (isEmpty(appId)) {
				System.out.println("Remove requested, but no app ID was provided.  Sending 404."); //$NON-NLS-1$
				return getIndexPage();
			}
			else {
				List fileList = (List) fileMap.get(appId);
				if (fileList != null) {
					dropPanel.removeFiles(appId, request.getParameter(APP_NAME), fileList);
					fileMap.remove(appId);
					String message = "Successfully removed " + //$NON-NLS-1$
							oneOrMore(fileList.size(), "file", "files") + //$NON-NLS-1$ //$NON-NLS-2$
							" from the list."; //$NON-NLS-1$

					return message;
				}
				else {
					return "Remove request received before list request.  Nothing done."; //$NON-NLS-1$
				}
			}
		}

		if (requestedFile.equals("/hash")) { //$NON-NLS-1$
			ParsedHTTPRequest newRequest = ParsedHTTPRequest.parseRequest(
					"/" + request.getParameter("file") + "?" + request.getCookedParams());

			InputStream is = getInputStream(newRequest);
			if (is == null) {
				return "The parameters did not refer to a valid file.";
			}

			try {
				MessageDigest md5 = MessageDigest.getInstance("MD5");
				byte[] inputFileInBytes = getInputFileAsBytes(is);
				byte[] digest = md5.digest(inputFileInBytes);
				is.close();

				return convertBytesToHexString(digest);
			}
			catch (IOException e) {
				return "Unable to compute hash:  cannot read file.";
			}
			catch (NoSuchAlgorithmException e) {
				return "Unable to compute hash:  MD5 not supported.";
			}
		}

		return null;
	}

	public static String convertBytesToHexString(byte[] array) {
		StringBuffer sb = new StringBuffer(array.length * 2);
		for(int x = 0 ; x < array.length ; x++) {
			String byteAsString = Integer.toHexString(array[x]).toUpperCase();
			if (byteAsString.length() == 1) {
				sb.append("0");
				sb.append(byteAsString);
			}
			else if (byteAsString.length() == 2) {
				sb.append(byteAsString);
			}
			else {
				sb.append(byteAsString.substring(byteAsString.length() - 2));
			}
		}

		return sb.toString();
	}

	private void transfer(InputStream is, OutputStream os) throws IOException {
		byte[] buffer = new byte[BUFFER_SIZE];

		while (true) {
			int len = is.read(buffer);
			if (len == -1) { break; }

			os.write(buffer, 0, len);
		}
	}

	private byte[] getInputFileAsBytes(InputStream is) throws IOException {
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		transfer(is, bos);
		return bos.toByteArray();
	}

	private String oneOrMore(int i, String singular, String plural) {
		if (i == 1) {
			return "1 " + singular; //$NON-NLS-1$
		}
		else {
			return i + " " + plural; //$NON-NLS-1$
		}
	}

	private String oneOrMoreWithPrefix(String singularPrefix, String pluralPrefix,
			int i, String singular, String plural)
	{
		if (i == 1) {
			return singularPrefix + "1" + singular; //$NON-NLS-1$
		}
		else {
			return pluralPrefix + i + plural; //$NON-NLS-1$
		}
	}

	public InputStream getInputStream(ParsedHTTPRequest request) {
		String handleBuiltins = handleBuiltins(request);
		if (handleBuiltins != null) {
			try {
				return new ByteArrayInputStream(handleBuiltins.getBytes("UTF-8")); //$NON-NLS-1$
			}
			catch (UnsupportedEncodingException e) {
				JOptionPane.showMessageDialog(null,
						messages.getString("SyncDocs.error.utf8NotAvailable"), //$NON-NLS-1$
						messages.getString("SyncDocs.title.utf8NotAvailable"), //$NON-NLS-1$
						JOptionPane.ERROR_MESSAGE);

				return null;
			}
		}

		String dir = request.getParameter(DIR);
		if (dir == null) {
			return getInputStreamFromRegularFileList(request);
		}
		else {
			File requestFile = validateRequestDir(request);
			if (requestFile == null) { return null; }

			try {
				return new FileInputStream(requestFile);
			}
			catch (FileNotFoundException e) {
				// This case can't happen because we've already checked for
				// file existence in the "canRead" method.
				e.printStackTrace();
				return null;
			}
		}
	}

	/**
	 * Get an InputStream for files that have been dropped onto the
	 * main window.
	 *
	 * @param request
	 * @return
	 */
	private InputStream getInputStreamFromRegularFileList(ParsedHTTPRequest request) {
		String requestedFile = request.getFilename();

		List filenameList = dropPanel.getFilenameList();
		Iterator filenameIter = filenameList.iterator();
		while (filenameIter.hasNext()) {
			FileEntry fileEntry = (FileEntry) filenameIter.next();

			String appId = request.getParameter(APP_ID);
			if (!fileEntry.validateAppId(appId)) {
				continue;
			}

			String filename = fileEntry.getFileName();
			File file = new File(filename);
			if (requestedFile.equals("/" + file.getName())) { //$NON-NLS-1$
				try {
					return new FileInputStream(file);
				}
				catch (FileNotFoundException e) {
					// TODO:  This case will happen if the file has been
					// renamed or moved.  It should be handled more gracefully.
					try {
						System.out.println("Unable to find file:  '" + file.getCanonicalPath() + "'.");
					}
					catch (IOException e1) {
						System.out.println("Unable to get the canonical path of:  '" + file.getAbsolutePath() + "'.");
					}

					return null;
				}
			}
		}

		return null;
	}

	private String getIndexPage() {
		StringBuffer sb = new StringBuffer();

		sb.append("<html><head><title>SyncDocs " + VERSION + ":  Retrieve File List</title></head><body>\r\n"); //$NON-NLS-1$ //$NON-NLS-2$
		sb.append("<form action=\"listWithDirs\" method=\"GET\">\r\n"); //$NON-NLS-1$
		sb.append("<table border='0'>"); //$NON-NLS-1$
		sb.append("<tr><td>Filetypes/extensions:</td><td><input type=\"text\" name=\"filetype\"></td></tr>"); //$NON-NLS-1$
		sb.append("<tr><td>App ID:</td><td><input type=\"text\" name=\"appId\"></td></tr>"); //$NON-NLS-1$
		sb.append("<tr><td>App Name:</td><td><input type=\"text\" name=\"appName\"></td></tr>"); //$NON-NLS-1$
		sb.append("<tr><td><input type=\"submit\" value=\"Submit\"></td></tr></table>\r\n"); //$NON-NLS-1$
		sb.append("</form>\r\n"); //$NON-NLS-1$
		sb.append("<ul>"); //$NON-NLS-1$
		sb.append("<li>Multiple filetypes/extensions can be provided as a comma separated list.</li>"); //$NON-NLS-1$
		sb.append("<li>App ID is only required if the requested files have been set to sync with a specific application.</li>"); //$NON-NLS-1$
		sb.append("<li>App Name is not required, but will display in SyncDocs instead of the App ID if provided.</li>"); //$NON-NLS-1$
		sb.append("</ul>"); //$NON-NLS-1$
		sb.append("</body></html>\r\n"); //$NON-NLS-1$

		return sb.toString();
	}

	private String getListPage(ParsedHTTPRequest request, String fileTypes)
	{
		String dir = request.getParameter(DIR);

		if (dir == null && fileTypes == null) {
			return getIndexPage();
		}

		StringBuffer sb = new StringBuffer();

		sb.append("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\r\n"); //$NON-NLS-1$
		sb.append("<html><head><title>SyncDocs " + VERSION + ":  Available Files</title></head><body>\r\n"); //$NON-NLS-1$ //$NON-NLS-2$


		boolean filesFound;
		if (dir == null) {
			filesFound = handleRegularFileList(request, fileTypes, sb);
		}
		else {
			filesFound = handleDirFileList(request, fileTypes, sb);
		}

		if (!filesFound) {
			sb.append("No files available."); //$NON-NLS-1$
		}

		sb.append("</body></html>\r\n"); //$NON-NLS-1$

		return sb.toString();
	}

	/**
	 * Handle returning the list of files that have been dropped onto the
	 * main window and any backup directories that have been made.
	 * 
	 * @param request
	 * @param fileTypes
	 * @param sb
	 * 
	 * @return
	 */
	private boolean handleRegularFileList(ParsedHTTPRequest request,
			String fileTypes, StringBuffer sb)
	{
		String appId = request.getParameter(APP_ID);

		List filenameList = dropPanel.getFilenameList();
		Iterator filenameIter = filenameList.iterator();
		boolean filesFound = false;

		String[] fileTypeArray = fileTypes.split(","); //$NON-NLS-1$
		Set fileTypeSet = new HashSet();

		for (int i = 0; i < fileTypeArray.length; i++) {
			if (fileTypeArray[i].length() > 0) {
				fileTypeSet.add(fileTypeArray[i].toLowerCase());
			}
		}

		while (filenameIter.hasNext()) {
			FileEntry fileEntry = (FileEntry) filenameIter.next();

			File file = new File(fileEntry.getFileName());
			if (!fileEntry.validateAppId(appId)) { continue; }
			if (!file.exists()) {
		        String message = messages.getString("SyncDocs.error.fileNoLongerAvailable1") +
        		fileEntry.getFileName() +
        		messages.getString("SyncDocs.error.fileNoLongerAvailable2");

				JOptionPane pane = new JOptionPane(message);
		        pane.setMessageType(JOptionPane.INFORMATION_MESSAGE);
		        JDialog dialog = pane.createDialog(appFrame, messages.getString("SyncDocs.title.fileNoLongerAvailable")); //$NON-NLS-1$
		        dialog.setModal(false);
		        dialog.setVisible(true);

				continue;
			}

			String filename = fileEntry.getFileName();
			String ext = getExtension(filename);
			if (!fileTypeSet.contains(ext)) { continue; }

			filesFound = true;

			fileMap.putElement(appId, filename);

			Date lastModified = new Date(file.lastModified());
			SimpleDateFormat formatter = new SimpleDateFormat(XML_DATE_FORMAT);
			String lastModifiedStr = formatter.format(lastModified);

			String encodedFileName = file.getName().replaceAll("&", "&amp;").replaceAll("<", "&lt;").replaceAll(">", "&gt;");
			sb.append("<a dir=\"false\" lastModifiedDate=\"" + lastModifiedStr + "\" href=\"" + encodedFileName + "\">" + encodedFileName + "</a><br />\r\n"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
		}

		if (request.getFilename().equals("/list")) {
			return filesFound;
		}

		File outputDir = validateRequestDir(request);
		if (outputDir == null) {
			return filesFound;
		}

		if (!outputDir.exists()) {
			//outputDir.mkdirs();
			return filesFound;
		}

		// Now check the downloads directory for any directories.
		File[] files = outputDir.listFiles();
		for (int i = 0; i < files.length; i++) {
			File file = files[i];

			if (file.isDirectory()) {
				Date lastModified = new Date(file.lastModified());
				SimpleDateFormat formatter = new SimpleDateFormat(XML_DATE_FORMAT);
				String lastModifiedStr = formatter.format(lastModified);

				sb.append("<a dir=\"true\" lastModifiedDate=\""); //$NON-NLS-1$
				sb.append(lastModifiedStr);
				sb.append("\" href=\"" + request.getCooked() + "&dir=" + file.getName() + "/\">"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
				sb.append(file.getName() + "</a><br>\r\n"); //$NON-NLS-1$

				filesFound = true;
			}
		}

		return filesFound;
	}

	/**
	 * Handle returning the backup directories that have been made.
	 * 
	 * @param request
	 * @param fileTypes
	 * @param sb
	 * 
	 * @return
	 */
	private boolean handleDirFileList(ParsedHTTPRequest request,
			String fileTypes, StringBuffer sb)
	{
		boolean filesFound = false;

		Set fileTypeSet = new HashSet();
		if (fileTypes != null) {
			String[] fileTypeArray = fileTypes.split(","); //$NON-NLS-1$
	
			for (int i = 0; i < fileTypeArray.length; i++) {
				if (fileTypeArray[i].length() > 0) {
					fileTypeSet.add(fileTypeArray[i].toLowerCase());
				}
			}
		}

		File backupDir = validateRequestDir(request);
		if (backupDir == null) {
			return filesFound;
		}

		if (!backupDir.exists()) {
			// CE 2008-12-07
			// Removed directory creation because it will create a new
			// directory if a listing is requested for a directory that
			// doesn't exist
			//backupDir.mkdirs();
			return filesFound;
		}

		String dir = request.getParameter(DIR);

		File[] files = backupDir.listFiles();
		for (int i = 0; i < files.length; i++) {
			File file = files[i];

			String filename = file.getName();
			String ext = getExtension(filename);
			if (!file.isDirectory()) {
				if (!fileTypeSet.isEmpty() && !fileTypeSet.contains(ext)) {
					continue;
				}
			}

			filesFound = true;

			Date lastModified = new Date(file.lastModified());
			SimpleDateFormat formatter = new SimpleDateFormat(XML_DATE_FORMAT);
			String lastModifiedStr = formatter.format(lastModified);

			ParsedHTTPRequest newRequest = request.copy();
			newRequest.removeParameter("dir");

			StringBuffer sb2 = new StringBuffer();
			if (!file.isDirectory()) {
				sb2.append("<a dir=\"false\" lastModifiedDate=\""); //$NON-NLS-1$
				sb2.append(lastModifiedStr);
				sb2.append("\" href=\"/"); //$NON-NLS-1$
				sb2.append(ParsedHTTPRequest.encode(file.getName()) + "?"); //$NON-NLS-1$
				newRequest.getCookedParams(sb2);
				sb2.append("&dir="); //$NON-NLS-1$
				sb2.append(dir);
				sb2.append("\">"); //$NON-NLS-1$
				sb2.append(file.getName() + "</a><br>\r\n"); //$NON-NLS-1$
			}
			else {
				sb2.append("<a dir=\"true\" lastModifiedDate=\""); //$NON-NLS-1$
				sb2.append(lastModifiedStr);
				sb2.append("\" href=\"" + newRequest.getCooked() + "&dir=" + dir + file.getName() + "/\">"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
				sb2.append(file.getName() + "</a><br>\r\n"); //$NON-NLS-1$
			}

			sb.append(sb2.toString().replaceAll("&", "&amp;"));
		}

		return filesFound;
	}

	private File validateRequestDir(ParsedHTTPRequest request) {
		String dir = request.getParameter(DIR);
		if (dir == null) {
			return null;
		}

		File downloadsDir = getDownloadsDirectory(request);
		File requestPath = new File(downloadsDir, dir);

		if (!isBuiltin(request)) {
			requestPath = new File(requestPath, request.getFilename());
		}

		try {
			String canonicalDownloadsPath = downloadsDir.getCanonicalPath();
			String canonicalRequestPath = requestPath.getCanonicalPath();

			// Check for ../ and other craziness.
			if (!canonicalRequestPath.startsWith(canonicalDownloadsPath)) {
				return null;
			}

			return requestPath;
		}
		catch (IOException e) {
			System.err.println("Unable to canonicalize path.");
			e.printStackTrace();
			return null;
		}
	}

	public File getDownloadsDirectory(ParsedHTTPRequest request) {
		File baseFile;
		if (useUserDirForDownloads) {
			baseFile = new File(System.getProperty("user.home"), "SyncDocs");
			if (baseFile == null) {
				JOptionPane.showMessageDialog(null,
						messages.getString("SyncDocs.error.userDirNotAvailable"), //$NON-NLS-1$
						messages.getString("SyncDocs.title.userDirNotAvailable"), //$NON-NLS-1$
						JOptionPane.ERROR_MESSAGE);

				useUserDirForDownloads = false;
				baseFile = new File("./downloads/");
			}
		}
		else {
			baseFile = new File("./downloads/");
		}

		if (request == null) {
			return baseFile;
		}

		//String appId = request.getParameter(APP_ID);
		String appName = request.getParameter(APP_NAME);
		if (appName != null) {
			return new File(baseFile, appName); //$NON-NLS-1$
		}
		else {
			return baseFile;
		}
	}

	public String getContentType(ParsedHTTPRequest request) {
		if (isBuiltin(request)) {
			return "text/html; charset=utf-8"; //$NON-NLS-1$
		}

		String requestedFile = request.getFilename();

		String ext = getExtension(requestedFile);
		if (ext == null) {
			return null;
		}

		String mimeType = (String) mimeTypes.get(ext);
		if (mimeType != null) {
			return mimeType;
		}

		return null;
	}

	public boolean isBuiltin(ParsedHTTPRequest request) {
		String requestedFile = request.getFilename();

		if (requestedFile.equals("/") || //$NON-NLS-1$
				requestedFile.equals("/index.html") || //$NON-NLS-1$
				requestedFile.equals("/list") || //$NON-NLS-1$
				requestedFile.equals("/listWithDirs") || //$NON-NLS-1$
				requestedFile.equals("/remove") || //$NON-NLS-1$
				requestedFile.equals("/hash")) //$NON-NLS-1$
		{
			return true;
		}

		return false;
	}

	public String getExtension(String filename) {
		int index = filename.lastIndexOf("."); //$NON-NLS-1$
		if (index == -1) {
			return null;
		}

		return filename.substring(index + 1).toLowerCase();
	}

	public void handleException(Exception e) {
		e.printStackTrace();

		JOptionPane.showMessageDialog(null,
				messages.getString("SyncDocs.error.unableToBindToBonjour"), //$NON-NLS-1$
				messages.getString("SyncDocs.title.unableToBindToBonjour"), //$NON-NLS-1$
				JOptionPane.ERROR_MESSAGE);
	}

	public void terminated(List inZeroConfThreads, InetAddress netAddress) {
		synchronized (inZeroConfThreads) {
			Iterator threadIter = inZeroConfThreads.iterator();
			while (threadIter.hasNext()) {
				ZeroConfThread thread = (ZeroConfThread) threadIter.next();
				if (safeEquals(thread.getInetAddress(), netAddress)) {
					threadIter.remove();
					inZeroConfThreads.notify();
				}
			}
		}
	}

	public String getLocalisedString(String key) {
		return messages.getString(key);
	}

	public Messages getMessages() {
		return messages;
	}

	public void propertyChange(PropertyChangeEvent event) {
		if (event.getSource().equals(messages)) {
			welcomeLabel.setText(messages.getString("SyncDocs.label.dragFilesBelow")); //$NON-NLS-1$
			languagesButton.setText(messages.getString("SyncDocs.button.languages")); //$NON-NLS-1$
			showDownloadsButton.setText(messages.getString("SyncDocs.button.showDownloads")); //$NON-NLS-1$
			setupButton.setText(messages.getString("SyncDocs.button.setup")); //$NON-NLS-1$
			syncButton.setText(messages.getString("SyncDocs.button.sync")); //$NON-NLS-1$
			removeButton.setText(messages.getString("SyncDocs.button.remove")); //$NON-NLS-1$
			helpButton.setText(messages.getString("SyncDocs.button.help")); //$NON-NLS-1$

			resizeToButtonPanel();
		}
	}

	private void resizeToButtonPanel() {
		Dimension preferredSize = buttonPanel.getPreferredSize();
		Dimension currentSize = appFrame.getSize();
		Dimension minSize = appFrame.getMinimumSize();

		if (minSize.width < preferredSize.width + INSET) {
			appFrame.setMinimumSize(new Dimension(preferredSize.width + INSET, minSize.height));
		}

		if (currentSize.width < preferredSize.width + INSET) {
			appFrame.setSize(new Dimension(preferredSize.width + INSET, currentSize.height));
		}
	}

}
