/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.syncdocs;

public class Version {

	private int major = 0;
	private int minor = 0;
	private int maintenance = 0;
	private String debugBuildId = null;

	public Version(String versionString) {
		String[] components = versionString.split("\\.");

		if (components.length != 2 && components.length != 3) {
			throw new IllegalArgumentException("Version number must have either one or two periods.");
		}

		major = Integer.parseInt(components[0]);
		if (components[1].indexOf("-") > -1) {
			String[] lastComponentSplit = components[1].split("-");

			minor = Integer.parseInt(lastComponentSplit[0]);
			debugBuildId = lastComponentSplit[1];
		}
		else {
			minor = Integer.parseInt(components[1]);
		}

		if (components.length == 3) {
			if (components[2].indexOf("-") > -1) {
				String[] lastComponentSplit = components[2].split("-");

				maintenance = Integer.parseInt(lastComponentSplit[0]);
				debugBuildId = lastComponentSplit[1];
			}
			else {
				maintenance = Integer.parseInt(components[2]);
			}
		}
	}

	public boolean isVersionAtLeast(String versionStr) {
		return isVersionAtLeast(new Version(versionStr));
	}

	public boolean isVersionAtLeast(Version version) {
		if (version.major > major) {
			return false;
		}

		if (version.major == major) {
			if (version.minor > minor) {
				return false;
			}

			if (version.minor == minor) {
				if (version.maintenance > maintenance) {
					return false;
				}
			}

			if (debugBuildId == null && version.debugBuildId != null) {
				return false;
			}

			if (debugBuildId != null && version.debugBuildId != null) {
				if (debugBuildId.compareTo(version.debugBuildId) == -1) {
					return false;
				}
			}
		}

		return true;
	}

	public String getDebugBuildId() {
		return debugBuildId;
	}

	public int getMaintenance() {
		return maintenance;
	}

	public int getMajor() {
		return major;
	}

	public int getMinor() {
		return minor;
	}

}
