/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.syncdocs;

import java.io.IOException;
import java.net.InetAddress;
import java.util.List;

public abstract class ZeroConfThread extends Thread {

	private boolean shuttingDown = false;
	private List zeroConfThreads;
	protected IExceptionCallback exCallback;
	private ITerminatedCallback termCallback;

	protected InetAddress iNetAddress;
	protected String serviceName;
	protected int port;

	public ZeroConfThread(InetAddress iNetAddress, String serviceName, int port,
			List zeroConfThreads, IExceptionCallback callback,
			ITerminatedCallback termCallback)
	{
		this.iNetAddress = iNetAddress;
		this.serviceName = serviceName;
		this.port = port;

		this.zeroConfThreads = zeroConfThreads;
		this.exCallback = callback;
		this.termCallback = termCallback;
	}

	public static boolean requiresOneThreadPerDomain() {
		return true;
	}

	protected void waitForShutdown() {
	    synchronized(this) {
	    	do {
		    	try {
		    		if (!shuttingDown) {
		    			wait();
		    		}
				}
		    	catch (InterruptedException e) {
		    		// Ignore it.
		    	}
			}
	    	while (!shuttingDown);

	    	System.out.println(Thread.currentThread().getName() + ":  Unbinding.");

	    	protectedUnbind();
	    	termCallback.terminated(zeroConfThreads, iNetAddress);

	    	System.out.println(Thread.currentThread().getName() + ":  Unbound.");
	    }
	}

	public void run() {
		try {
			bind();
		    waitForShutdown();
		}
		catch (Exception e) {
			exCallback.handleException(e);
		}
	}

	public abstract void bind() throws IOException;

	public synchronized void unbind() {
		shuttingDown = true;
		notify();
	}

	protected abstract void protectedUnbind();

	public InetAddress getInetAddress() {
		return iNetAddress;
	}

}
