/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.syncdocs.test;

import ca.quine.syncdocs.Version;
import junit.framework.TestCase;

public class VersionTest extends TestCase {

	public void testCreateSimpleVersion() {
		Version v = new Version("1.2");
		assertEquals(1, v.getMajor());
		assertEquals(2, v.getMinor());
	}

	public void testCreateSimpleVersionWithBuildId() {
		Version v = new Version("1.2-RC1");
		assertEquals(1, v.getMajor());
		assertEquals(2, v.getMinor());
		assertEquals(0, v.getMaintenance());
		assertEquals("RC1", v.getDebugBuildId());
	}

	public void testCreateVersion() {
		Version v = new Version("1.2.3");
		assertEquals(1, v.getMajor());
		assertEquals(2, v.getMinor());
		assertEquals(3, v.getMaintenance());
	}

	public void testCreateVersionWithBuildId() {
		Version v = new Version("1.2.3-RC1");
		assertEquals(1, v.getMajor());
		assertEquals(2, v.getMinor());
		assertEquals(3, v.getMaintenance());
		assertEquals("RC1", v.getDebugBuildId());
	}

	public void testVersion() {
		assertTrue(new Version("0.1.0").isVersionAtLeast("0.0.9"));
		assertTrue(new Version("0.1.0").isVersionAtLeast("0.1.0"));
		assertTrue(new Version("0.1.1").isVersionAtLeast("0.1.0"));
		assertTrue(new Version("0.1.1").isVersionAtLeast("0.1.1"));

		assertFalse(new Version("0.1.2").isVersionAtLeast("0.1.3"));
		assertFalse(new Version("0.0.9").isVersionAtLeast("0.1.0"));
		assertFalse(new Version("0.1.0").isVersionAtLeast("0.1.1"));

		assertTrue(new Version("0.1.0-RC1").isVersionAtLeast("0.1.0"));
		assertFalse(new Version("0.1.0").isVersionAtLeast("0.1.0-RC1"));

		assertTrue(new Version("0.1.2-RC2").isVersionAtLeast("0.1.2-RC1"));
		assertTrue(new Version("0.1.2-RC2").isVersionAtLeast("0.1.2-RC2"));

		assertFalse(new Version("0.1.2-RC1").isVersionAtLeast("0.1.2-RC2"));
	}

}
