/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.web;

import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

public class ParsedHTTPRequest {

	private String cookedRequest;
	private String filename;
	private HashMap parameters;

	private ParsedHTTPRequest(String cookedRequest) {
		this.cookedRequest = cookedRequest;
		parameters = new HashMap();
	}

	public static ParsedHTTPRequest parseRequest(String cookedRequest) {
		ParsedHTTPRequest retVal = new ParsedHTTPRequest(cookedRequest);

		int questionIndex = cookedRequest.indexOf("?");
		if (questionIndex == -1) {
			retVal.filename = decode(cookedRequest);
			return retVal;
		}

		retVal.filename = decode(cookedRequest.substring(0, questionIndex));

		String cookedParameters = cookedRequest.substring(questionIndex + 1);
		String[] parameterPairs = cookedParameters.split("&");
		for (int i = 0; i < parameterPairs.length; i++) {
			String paramPair = parameterPairs[i];
			String[] keyValue = paramPair.split("=", 2);
			if (keyValue.length == 1) {
				retVal.parameters.put(decode(keyValue[0]), "");
			}
			else {
				retVal.parameters.put(decode(keyValue[0]), decode(keyValue[1]));
			}
		}

		return retVal;
	}

	public String getFilename() {
		return filename;
	}

	public String toString() {
		return cookedRequest;
	}

	public boolean hasParameters() {
		return parameters.size() > 0;
	}

	public String getParameter(String paramKey) {
		return (String) parameters.get(paramKey);
	}

	/**
	 * Take the raw request and "cook" it back into it's original form.
	 * @return
	 */
	public String getCooked() {
		StringBuffer sb = new StringBuffer();

		// Don't encode the filename because we want to return the initial
		// slash as is.
		sb.append(filename);
		sb.append("?");

		getCookedParams(sb);

		return sb.toString();
	}

	public String getCookedParams() {
		StringBuffer sb = new StringBuffer();
		getCookedParams(sb);

		return sb.toString();
	}

	public void getCookedParams(StringBuffer sb) {
		Iterator entrySetIter = parameters.entrySet().iterator();
		while (entrySetIter.hasNext()) {
			Map.Entry entry = (Map.Entry) entrySetIter.next();
			sb.append(encode((String) entry.getKey()));
			sb.append("=");
			sb.append(encode((String) entry.getValue()));

			if (entrySetIter.hasNext()) {
				sb.append("&");
			}
		}
	}

	private static String decode(String input) {
		try {
			return URLDecoder.decode(input, "UTF-8");
		}
		catch (UnsupportedEncodingException e) {
			throw new IllegalStateException("Fatal error:  UTF-8 encoding not supported");
		}
	}

	/**
	 * Encodes the input String with percent escapes in UTF-8
	 * encoding (the result is suitable for URLs).
	 * This method is public for SyncDocs.
	 * 
	 * @param input
	 * @return
	 */
	public static String encode(String input) {
		try {
			return URLEncoder.encode(input, "UTF-8");
		}
		catch (UnsupportedEncodingException e) {
			throw new IllegalStateException("Fatal error:  UTF-8 encoding not supported");
		}
	}

	public ParsedHTTPRequest copy() {
		ParsedHTTPRequest retVal = parseRequest(cookedRequest);

		return retVal;
	}

	public void removeParameter(String key) {
		parameters.remove(key);
		cookedRequest = getCooked();
	}

}
